"""
Minion data cache plugin for Consul key/value data store.

.. versionadded:: 2016.11.2

.. versionchanged:: 3005

    Timestamp/cache updated support added.

:depends: python-consul >= 0.2.0

It is up to the system administrator to set up and configure the Consul
infrastructure. All is needed for this plugin is a working Consul agent
with a read-write access to the key-value store.

The related documentation can be found in the `Consul documentation`_.

To enable this cache plugin, the master will need the python client for
Consul installed. This can be easily installed with pip:

.. code-block:: bash

    pip install python-consul

Optionally, depending on the Consul agent configuration, the following values
could be set in the master config. These are the defaults:

.. code-block:: yaml

    consul.host: 127.0.0.1
    consul.port: 8500
    consul.token: None
    consul.scheme: http
    consul.consistency: default
    consul.dc: dc1
    consul.verify: True
    consul.timestamp_suffix: .tstamp  # Added in 3005.0

In order to bring the cache APIs into conformity, in 3005.0 timestamp
information gets stored as a separate ``{key}.tstamp`` key/value. If your
existing functionality depends on being able to store normal keys with the
``.tstamp`` suffix, override the ``consul.timestamp_suffix`` default config.

Related docs could be found in the `python-consul documentation`_.

To use the consul as a minion data cache backend, set the master ``cache`` config
value to ``consul``:

.. code-block:: yaml

    cache: consul


.. _`Consul documentation`: https://www.consul.io/docs/index.html
.. _`python-consul documentation`: https://python-consul.readthedocs.io/en/latest/#consul

"""

import logging
import time

import salt.payload
from salt.exceptions import SaltCacheError

try:
    import consul

    HAS_CONSUL = True
except ImportError:
    HAS_CONSUL = False


log = logging.getLogger(__name__)
api = None
_tstamp_suffix = ".tstamp"


# Define the module's virtual name
__virtualname__ = "consul"

__func_alias__ = {"list_": "list"}


def __virtual__():
    """
    Confirm this python-consul package is installed
    """
    if not HAS_CONSUL:
        return (
            False,
            "Please install python-consul package to use consul data cache driver",
        )

    consul_kwargs = {
        "host": __opts__.get("consul.host", "127.0.0.1"),
        "port": __opts__.get("consul.port", 8500),
        "token": __opts__.get("consul.token", None),
        "scheme": __opts__.get("consul.scheme", "http"),
        "consistency": __opts__.get("consul.consistency", "default"),
        "dc": __opts__.get("consul.dc", None),
        "verify": __opts__.get("consul.verify", True),
    }

    try:
        global api, _tstamp_suffix
        _tstamp_suffix = __opts__.get("consul.timestamp_suffix", _tstamp_suffix)
        api = consul.Consul(**consul_kwargs)
    except AttributeError:
        return (
            False,
            "Failed to invoke consul.Consul, please make sure you have python-consul >="
            " 0.2.0 installed",
        )

    return __virtualname__


def store(bank, key, data):
    """
    Store a key value.
    """
    c_key = f"{bank}/{key}"
    tstamp_key = f"{bank}/{key}{_tstamp_suffix}"

    try:
        c_data = salt.payload.dumps(data)
        api.kv.put(c_key, c_data)
        api.kv.put(tstamp_key, salt.payload.dumps(int(time.time())))
    except Exception as exc:  # pylint: disable=broad-except
        raise SaltCacheError(f"There was an error writing the key, {c_key}: {exc}")


def fetch(bank, key):
    """
    Fetch a key value.
    """
    c_key = f"{bank}/{key}"
    try:
        _, value = api.kv.get(c_key)
        if value is None:
            return {}
        return salt.payload.loads(value["Value"])
    except Exception as exc:  # pylint: disable=broad-except
        raise SaltCacheError(f"There was an error reading the key, {c_key}: {exc}")


def flush(bank, key=None):
    """
    Remove the key from the cache bank with all the key content.
    """
    if key is None:
        c_key = bank
        tstamp_key = None
    else:
        c_key = f"{bank}/{key}"
        tstamp_key = f"{bank}/{key}{_tstamp_suffix}"
    try:
        if tstamp_key:
            api.kv.delete(tstamp_key)
        return api.kv.delete(c_key, recurse=key is None)
    except Exception as exc:  # pylint: disable=broad-except
        raise SaltCacheError(f"There was an error removing the key, {c_key}: {exc}")


def list_(bank):
    """
    Return an iterable object containing all entries stored in the specified bank.
    """
    try:
        _, keys = api.kv.get(bank + "/", keys=True, separator="/")
    except Exception as exc:  # pylint: disable=broad-except
        raise SaltCacheError(f'There was an error getting the key "{bank}": {exc}')
    if keys is None:
        keys = []
    else:
        # Any key could be a branch and a leaf at the same time in Consul
        # so we have to return a list of unique names only.
        out = set()
        for key in keys:
            out.add(key[len(bank) + 1 :].rstrip("/"))
        keys = [o for o in out if not o.endswith(_tstamp_suffix)]
    return keys


def contains(bank, key):
    """
    Checks if the specified bank contains the specified key.
    """
    try:
        c_key = "{}/{}".format(bank, key or "")
        _, value = api.kv.get(c_key, keys=True)
    except Exception as exc:  # pylint: disable=broad-except
        raise SaltCacheError(f"There was an error getting the key, {c_key}: {exc}")
    return value is not None


def updated(bank, key):
    """
    Return the Unix Epoch timestamp of when the key was last updated. Return
    None if key is not found.
    """
    c_key = f"{bank}/{key}{_tstamp_suffix}"
    try:
        _, value = api.kv.get(c_key)
        if value is None:
            return None
        return salt.payload.loads(value["Value"])
    except Exception as exc:  # pylint: disable=broad-except
        raise SaltCacheError(f"There was an error reading the key, {c_key}: {exc}")



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

import typing

from cryptography import utils
from cryptography.exceptions import (
    AlreadyFinalized,
    InvalidKey,
    UnsupportedAlgorithm,
    _Reasons,
)
from cryptography.hazmat.bindings._rust import openssl as rust_openssl
from cryptography.hazmat.primitives import constant_time, hashes
from cryptography.hazmat.primitives.kdf import KeyDerivationFunction


class PBKDF2HMAC(KeyDerivationFunction):
    def __init__(
        self,
        algorithm: hashes.HashAlgorithm,
        length: int,
        salt: bytes,
        iterations: int,
        backend: typing.Any = None,
    ):
        from cryptography.hazmat.backends.openssl.backend import (
            backend as ossl,
        )

        if not ossl.pbkdf2_hmac_supported(algorithm):
            raise UnsupportedAlgorithm(
                f"{algorithm.name} is not supported for PBKDF2.",
                _Reasons.UNSUPPORTED_HASH,
            )
        self._used = False
        self._algorithm = algorithm
        self._length = length
        utils._check_bytes("salt", salt)
        self._salt = salt
        self._iterations = iterations

    def derive(self, key_material: utils.Buffer) -> bytes:
        if self._used:
            raise AlreadyFinalized("PBKDF2 instances can only be used once.")
        self._used = True

        return rust_openssl.kdf.derive_pbkdf2_hmac(
            key_material,
            self._algorithm,
            self._salt,
            self._iterations,
            self._length,
        )

    def verify(self, key_material: bytes, expected_key: bytes) -> None:
        derived_key = self.derive(key_material)
        if not constant_time.bytes_eq(derived_key, expected_key):
            raise InvalidKey("Keys do not match.")



# This file is licensed under the BSD 2-Clause License.
# See https://opensource.org/licenses/BSD-2-Clause for details.

from ._curve import _Curve
from Cryptodome.Math.Numbers import Integer
from Cryptodome.Util._raw_api import (load_pycryptodome_raw_lib, VoidPointer,
                                  SmartPointer)


def curve25519_curve():
    p = 0x7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffed  # 2**255 - 19
    order = 0x1000000000000000000000000000000014def9dea2f79cd65812631a5cf5d3ed

    _curve25519_lib = load_pycryptodome_raw_lib("Cryptodome.PublicKey._curve25519", """
typedef void Point;

int curve25519_new_point(Point **out,
                         const uint8_t x[32],
                         size_t modsize,
                         const void* context);
int curve25519_clone(Point **P, const Point *Q);
void curve25519_free_point(Point *p);
int curve25519_get_x(uint8_t *xb, size_t modsize, Point *p);
int curve25519_scalar(Point *P, const uint8_t *scalar, size_t scalar_len, uint64_t seed);
int curve25519_cmp(const Point *ecp1, const Point *ecp2);
""")

    class EcLib(object):
        new_point = _curve25519_lib.curve25519_new_point
        clone = _curve25519_lib.curve25519_clone
        free_point = _curve25519_lib.curve25519_free_point
        get_x = _curve25519_lib.curve25519_get_x
        scalar = _curve25519_lib.curve25519_scalar
        cmp = _curve25519_lib.curve25519_cmp

    def _validate_x25519_point(point):

        p2 = p * 2
        x1 = 325606250916557431795983626356110631294008115727848805560023387167927233504
        x2 = 39382357235489614581723060781553021112529911719440698176882885853963445705823

        # http://cr.yp.to/ecdh.html#validate
        deny_list = (
            0,
            1,
            x1,
            x2,
            p - 1,
            p,
            p + 1,
            p + x1,
            p + x2,
            p2 - 1,
            p2,
            p2 + 1,
        )

        try:
            valid = point.x not in deny_list
        except ValueError:
            valid = False

        if not valid:
            raise ValueError("Invalid Curve25519 public key")

    curve25519 = _Curve(Integer(p),
                        None,
                        Integer(order),
                        Integer(9),
                        None,
                        None,
                        255,
                        "1.3.101.110",      # RFC8410
                        None,
                        "Curve25519",
                        None,
                        EcLib,
                        _validate_x25519_point,
                        )

    return curve25519


def curve448_curve():
    p = 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffffffffffffffffffffffffffffffffffffffffffffffffffff  # 2**448 - 2**224 - 1
    order = 0x3fffffffffffffffffffffffffffffffffffffffffffffffffffffff7cca23e9c44edb49aed63690216cc2728dc58f552378c292ab5844f3

    _curve448_lib = load_pycryptodome_raw_lib("Cryptodome.PublicKey._curve448", """
typedef void Curve448Context;
typedef void Curve448Point;

int curve448_new_context(Curve448Context **pec_ctx);
void curve448_free_context(Curve448Context *ec_ctx);
int curve448_new_point(Curve448Point **out,
                       const uint8_t *x,
                       size_t len,
                       const Curve448Context *ec_ctx);
void curve448_free_point(Curve448Point *p);
int curve448_clone(Curve448Point **P, const Curve448Point *Q);
int curve448_get_x(uint8_t *xb, size_t modsize, const Curve448Point *p);
int curve448_scalar(Curve448Point *P, const uint8_t *scalar, size_t scalar_len, uint64_t seed);
int curve448_cmp(const Curve448Point *ecp1, const Curve448Point *ecp2);
""")

    class EcLib(object):
        new_context = _curve448_lib.curve448_new_context
        free_context = _curve448_lib.curve448_free_context
        new_point = _curve448_lib.curve448_new_point
        clone = _curve448_lib.curve448_clone
        free_point = _curve448_lib.curve448_free_point
        get_x = _curve448_lib.curve448_get_x
        scalar = _curve448_lib.curve448_scalar
        cmp = _curve448_lib.curve448_cmp

    curve448_context = VoidPointer()
    result = EcLib.new_context(curve448_context.address_of())
    if result:
        raise ImportError("Error %d initializing Curve448 context" % result)

    def _validate_x448_point(point):
        deny_list = (
            0,
            1,
            p - 1,
            p,
            p + 1,
        )

        try:
            valid = point.x not in deny_list
        except ValueError:
            valid = False

        if not valid:
            raise ValueError("Invalid Curve448 public key")

    curve448 = _Curve(Integer(p),
                      None,
                      Integer(order),
                      Integer(5),
                      None,
                      None,
                      448,
                      "1.3.101.111",      # RFC8410
                      SmartPointer(curve448_context.get(), EcLib.free_context),
                      "Curve448",
                      None,
                      EcLib,
                      _validate_x448_point,
                      )

    return curve448



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

BROWSER_BRANDS: Final = "browser.brands"
"""
Array of brand name and version separated by a space.
Note: This value is intended to be taken from the [UA client hints API](https://wicg.github.io/ua-client-hints/#interface) (`navigator.userAgentData.brands`).
"""

BROWSER_LANGUAGE: Final = "browser.language"
"""
Preferred language of the user using the browser.
Note: This value is intended to be taken from the Navigator API `navigator.language`.
"""

BROWSER_MOBILE: Final = "browser.mobile"
"""
A boolean that is true if the browser is running on a mobile device.
Note: This value is intended to be taken from the [UA client hints API](https://wicg.github.io/ua-client-hints/#interface) (`navigator.userAgentData.mobile`). If unavailable, this attribute SHOULD be left unset.
"""

BROWSER_PLATFORM: Final = "browser.platform"
"""
The platform on which the browser is running.
Note: This value is intended to be taken from the [UA client hints API](https://wicg.github.io/ua-client-hints/#interface) (`navigator.userAgentData.platform`). If unavailable, the legacy `navigator.platform` API SHOULD NOT be used instead and this attribute SHOULD be left unset in order for the values to be consistent.
The list of possible values is defined in the [W3C User-Agent Client Hints specification](https://wicg.github.io/ua-client-hints/#sec-ch-ua-platform). Note that some (but not all) of these values can overlap with values in the [`os.type` and `os.name` attributes](./os.md). However, for consistency, the values in the `browser.platform` attribute should capture the exact value that the user agent provides.
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

HEROKU_APP_ID: Final = "heroku.app.id"
"""
Unique identifier for the application.
"""

HEROKU_RELEASE_COMMIT: Final = "heroku.release.commit"
"""
Commit hash for the current release.
"""

HEROKU_RELEASE_CREATION_TIMESTAMP: Final = "heroku.release.creation_timestamp"
"""
Time and date the release was created.
"""



import logging
import re

import salt.modules.cmdmod
import salt.utils.json
from salt.exceptions import CommandExecutionError

# Standard Salt logging
log = logging.getLogger(__name__)

HAS_CLR = False
try:
    import clr

    HAS_CLR = True
except ImportError:
    log.debug("could not import clr, pythonnet is not available")

HAS_PWSH_SDK = False
try:
    PWSH_GAC_NAME = "System.Management.Automation, Version=3.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"
    clr.AddReference(PWSH_GAC_NAME)
    from System.Management.Automation import PowerShell

    HAS_PWSH_SDK = True
except Exception as exc:  # pylint: disable=broad-exception-caught
    log.debug("win_pwsh could not load PowerShell SDK: %s", exc)


def run_dict(cmd, cwd=None):
    """
    Execute the PowerShell command and return the data as a dictionary

    .. versionadded:: 3006.9

    Args:

        cmd (str,list): The PowerShell command to run

        cwd (str): The current working directory

    Returns:
        dict: A dictionary containing the output of the PowerShell command

    Raises:
        CommandExecutionError:
            If an error is encountered or the command does not complete
            successfully
    """
    if isinstance(cmd, list):
        cmd = " ".join(map(str, cmd))
    if "convertto-json" not in cmd.lower():
        cmd = f"{cmd} | ConvertTo-Json"
    if "progresspreference" not in cmd.lower():
        cmd = f"$ProgressPreference = 'SilentlyContinue'; {cmd}"
    ret = salt.modules.cmdmod.run_all(cmd=cmd, shell="powershell", cwd=cwd)

    if "pid" in ret:
        del ret["pid"]

    if ret.get("stderr", ""):
        error = ret["stderr"].splitlines()[0]
        raise CommandExecutionError(error, info=ret)

    if "retcode" not in ret or ret["retcode"] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError("Issue executing PowerShell cmd", info=ret)

    # Sometimes Powershell returns an empty string, which isn't valid JSON
    if ret["stdout"] == "":
        ret["stdout"] = "{}"

    try:
        ret = salt.utils.json.loads(ret["stdout"], strict=False)
    except ValueError:
        raise CommandExecutionError("No JSON results from PowerShell", info=ret)

    return ret


class PowerShellSession:
    """
    A stateful PowerShell runspace backed by the .NET PowerShell SDK
    (``System.Management.Automation``), loaded via ``pythonnet``.

    Unlike :func:`run_dict`, which shells out to a ``powershell.exe``
    subprocess, this class opens an **in-process** runspace and keeps it alive
    across multiple calls.  The same runspace is reused for every ``run*``
    invocation, so imported modules, loaded functions, and session variables
    persist for the lifetime of the object.

    **Requirements:** both :data:`HAS_CLR` and :data:`HAS_PWSH_SDK` must be
    ``True`` before instantiating this class.

    **Usage — always use as a context manager** so the underlying runspace is
    disposed of deterministically::

        with PowerShellSession() as session:
            result = session.run_json("Get-NetAdapter | Select-Object Name, Status")

    **Methods:**

    * :meth:`run` — run a script and return raw Python scalars/lists.
    * :meth:`run_json` — run a script and return a parsed Python object via
      ``ConvertTo-Json``.
    * :meth:`run_strict` — like :meth:`run` but raises
      :exc:`~salt.exceptions.CommandExecutionError` if the script did not run
      to completion (i.e., an uncaught error occurred).

    **Session defaults** set in :meth:`__init__`:

    * ``$ErrorActionPreference = 'Stop'`` — all non-terminating cmdlet errors
      are promoted to terminating exceptions so they cannot be silently ignored.
    * ``$ProgressPreference = 'SilentlyContinue'`` — suppresses progress bars
      that would otherwise pollute the output stream.
    * ``$WarningPreference = 'SilentlyContinue'`` — suppresses advisory
      warnings that are not actionable in an automation context.
    """

    def __init__(self):
        """
        Create the PowerShell runspace and apply session-wide preferences.

        Sets ``$ErrorActionPreference = 'Stop'`` so every cmdlet uses
        terminating-error semantics by default.  Individual commands that are
        expected to return nothing (e.g., ``Get-NetRoute`` when no route
        exists) must use ``-ErrorAction SilentlyContinue`` or be wrapped in a
        PowerShell ``try/catch`` block.
        """
        # Create PowerShell instance
        self.ps = PowerShell.Create()

        # Suppress anything that might be displayed
        self.ps.AddScript("$ProgressPreference = 'SilentlyContinue'").AddStatement()
        self.ps.AddScript("$ErrorActionPreference = 'Stop'").AddStatement()
        self.ps.AddScript("$WarningPreference = 'SilentlyContinue'").AddStatement()
        self.ps.Invoke()

    def __enter__(self):
        return self

    def version(self):
        return self.run("$PSVersionTable.PSVersion.ToString()")

    def import_modules(self, modules=None):
        """
        Load modules into the existing session
        """
        if not modules:
            return

        self.ps.Commands.Clear()
        self.ps.Streams.ClearStreams()
        for module in modules if isinstance(modules, list) else [modules]:
            self.ps.AddScript(f"Import-Module {module}").AddStatement()
        self.ps.Invoke()

    def run(self, cmd):
        """
        Run a PowerShell command and return the result without attempting to
        parse it convert the PowerShell objects.

        Args:

            cmd (str): The command to run.

        Returns:
            str: A string with the results of the command
            list: If there is more than one return, it will be a list of strings
        """
        # Clear previous commands and any accumulated stream state
        self.ps.Commands.Clear()
        self.ps.Streams.ClearStreams()
        self.ps.AddScript(cmd)
        results = self.ps.Invoke()

        if self.ps.HadErrors:
            error_msg = "Unknown PowerShell Error"

            if self.ps.Streams.Error.Count > 0:
                error_msg = self.ps.Streams.Error[0].ToString()
            elif self.ps.InvocationStateInfo.Reason:
                error_msg = str(self.ps.InvocationStateInfo.Reason.Message)

            # We don't raise here so the session stays alive, but we log it
            log.debug("PowerShell Session Warning/Error: %s", error_msg)
            return error_msg

        if not results or len(results) == 0:
            return None

        if len(results) == 1:
            value = results[0]
            if value is None:
                return None

            # Access the underlying .NET BaseObject (e.g., System.String -> str)
            base_value = value.BaseObject

            # Basic type mapping
            if isinstance(base_value, (str, int, bool, float)):
                return base_value

            return str(base_value)

        # If there are multiple results, return them as a list of strings
        return [str(result) for result in results]

    def run_json(self, cmd, depth=4):
        """
        Run a PowerShell command and return the result as a parsed Python object.

        Unless the command already contains ``ConvertTo-Json``, the method
        automatically pipes the output through ``ConvertTo-Json -Compress
        -Depth <depth>`` and then ``Out-String`` before parsing. This ensures
        that PowerShell objects are always serialized to English-language JSON
        regardless of the system locale.

        Args:

            cmd (str): The PowerShell command or script to run.

            depth (int): The JSON serialization depth passed to
                ``ConvertTo-Json``. Defaults to ``4``.

        Returns:
            Any: The deserialized Python object (dict, list, str, etc.), or
            ``None`` if PowerShell produced no output.

        Raises:
            CommandExecutionError: If PowerShell enters a ``Failed`` state.
        """
        # Clear previous commands and any accumulated stream state
        self.ps.Commands.Clear()
        self.ps.Streams.ClearStreams()

        self.ps.AddScript(cmd)

        # Only add ConvertTo-Json if not already present in the command.
        # Match both piped (| ConvertTo-Json) and standalone (ConvertTo-Json
        # -InputObject ...) forms so we never double-encode the output.
        if not re.search(r"ConvertTo-Json", cmd, re.IGNORECASE):
            self.ps.AddCommand("ConvertTo-Json")
            self.ps.AddParameter("Compress", True)
            self.ps.AddParameter("Depth", depth)

        # Use Out-String for reliable string extraction from the pipeline
        if not cmd.lower().endswith("out-string"):
            self.ps.AddCommand("Out-String")

        results = self.ps.Invoke()

        if self.ps.HadErrors:
            if self.ps.InvocationStateInfo.State.ToString() == "Failed":
                err_msgs = [str(err) for err in self.ps.Streams.Error]
                raise CommandExecutionError(f"PowerShell Error: {err_msgs}")

        if results and len(results) > 0:
            return salt.utils.json.loads(str(results[0]))
        return None

    def run_strict(self, cmd):
        """
        Run a PowerShell command and raise on any error.

        Identical to :meth:`run` except that a ``CommandExecutionError`` is
        raised when PowerShell reports ``HadErrors``, rather than logging and
        returning the error string. Use this when the caller must not silently
        continue after a failure (e.g., when executing a large configuration
        script where a mid-script error would leave the system in an unknown
        state).

        .. note::
            PowerShell sets ``HadErrors = True`` even for errors caught by
            ``try/catch`` inside the script.  To distinguish "script ran to
            completion with some suppressed errors" from "script terminated
            mid-way", we bracket the caller's script with a sentinel variable.
            If the sentinel is ``$true`` at the end, all errors were caught
            intentionally and we do not raise.

        Args:

            cmd (str): The PowerShell command or script to run.

        Returns:
            str | list | None: Same return types as :meth:`run`.

        Raises:
            CommandExecutionError: If PowerShell reports any error.
        """
        self.ps.Commands.Clear()
        self.ps.Streams.ClearStreams()
        # Sentinel: set $false before the script, $true after.  If the script
        # terminates mid-way (uncaught error) the sentinel stays $false.
        wrapped = "$__salt_ok__ = $false\n" + cmd + "\n$__salt_ok__ = $true"
        self.ps.AddScript(wrapped)
        results = self.ps.Invoke()

        if self.ps.HadErrors:
            # Capture error info BEFORE the sentinel query overwrites state.
            streams_err = (
                self.ps.Streams.Error[0].ToString()
                if self.ps.Streams.Error.Count > 0
                else None
            )
            inv_reason = self.ps.InvocationStateInfo.Reason

            # Did the script run to completion?  If yes, every error was caught
            # by an explicit try/catch inside the script — do not raise.
            self.ps.Commands.Clear()
            self.ps.Streams.ClearStreams()
            self.ps.AddScript("$__salt_ok__ -eq $true")
            chk = self.ps.Invoke()
            ran_to_end = chk and chk.Count > 0 and str(chk[0]).lower() == "true"
            if ran_to_end:
                # All errors were intentionally suppressed; script completed.
                pass
            else:
                # Script terminated before the sentinel — uncaught error.
                error_msg = "Unknown PowerShell Error"
                if streams_err:
                    error_msg = streams_err
                elif inv_reason:
                    error_msg = str(inv_reason.Message)
                else:
                    # Terminating errors from $ErrorActionPreference = 'Stop'
                    # do not always populate Streams.Error in PS 5.1 SDK.
                    # Query $Error[0] from the runspace as a last-resort.
                    self.ps.Commands.Clear()
                    self.ps.AddScript(
                        "if ($Error.Count -gt 0)"
                        " { $Error[0].Exception.Message } else { '' }"
                    )
                    err_results = self.ps.Invoke()
                    if err_results and err_results.Count > 0:
                        msg = str(err_results[0])
                        if msg.strip():
                            error_msg = msg
                raise CommandExecutionError(error_msg)

        if not results or len(results) == 0:
            return None

        if len(results) == 1:
            value = results[0]
            if value is None:
                return None
            base_value = value.BaseObject
            if isinstance(base_value, (str, int, bool, float)):
                return base_value
            return str(base_value)

        return [str(result) for result in results]

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        exc_type: The class of the exception (e.g., CommandExecutionError)
        exc_val: The instance of the exception
        exc_tb: The traceback object
        """
        if exc_type:
            log.debug("PowerShellSession exiting due to error: %s", exc_val)
            log.debug(exc_tb)
        self.ps.Dispose()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar as _ClassVar
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import message as _message

DESCRIPTOR: _descriptor.FileDescriptor

class DateTime(_message.Message):
    __slots__ = (
        "year",
        "month",
        "day",
        "hours",
        "minutes",
        "seconds",
        "nanos",
        "utc_offset",
        "time_zone",
    )
    YEAR_FIELD_NUMBER: _ClassVar[int]
    MONTH_FIELD_NUMBER: _ClassVar[int]
    DAY_FIELD_NUMBER: _ClassVar[int]
    HOURS_FIELD_NUMBER: _ClassVar[int]
    MINUTES_FIELD_NUMBER: _ClassVar[int]
    SECONDS_FIELD_NUMBER: _ClassVar[int]
    NANOS_FIELD_NUMBER: _ClassVar[int]
    UTC_OFFSET_FIELD_NUMBER: _ClassVar[int]
    TIME_ZONE_FIELD_NUMBER: _ClassVar[int]
    year: int
    month: int
    day: int
    hours: int
    minutes: int
    seconds: int
    nanos: int
    utc_offset: _duration_pb2.Duration
    time_zone: TimeZone
    def __init__(
        self,
        year: _Optional[int] = ...,
        month: _Optional[int] = ...,
        day: _Optional[int] = ...,
        hours: _Optional[int] = ...,
        minutes: _Optional[int] = ...,
        seconds: _Optional[int] = ...,
        nanos: _Optional[int] = ...,
        utc_offset: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        time_zone: _Optional[_Union[TimeZone, _Mapping]] = ...,
    ) -> None: ...

class TimeZone(_message.Message):
    __slots__ = ("id", "version")
    ID_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    id: str
    version: str
    def __init__(
        self, id: _Optional[str] = ..., version: _Optional[str] = ...
    ) -> None: ...



"""
Functions used to sync external modules
"""

import logging
import os
import shutil

import salt.fileclient
import salt.utils.files
import salt.utils.hashutils
import salt.utils.path
import salt.utils.url
from salt.config import DEFAULT_HASH_TYPE

log = logging.getLogger(__name__)


def _list_emptydirs(rootdir):
    emptydirs = []
    for root, dirs, files in salt.utils.path.os_walk(rootdir):
        if not files and not dirs:
            emptydirs.append(root)
    return emptydirs


def _listdir_recursively(rootdir):
    file_list = []
    for root, dirs, files in salt.utils.path.os_walk(rootdir):
        for filename in files:
            relpath = os.path.relpath(root, rootdir).strip(".")
            file_list.append(os.path.join(relpath, filename))
    return file_list


def sync(
    opts,
    form,
    saltenv=None,
    extmod_whitelist=None,
    extmod_blacklist=None,
):
    """
    Sync custom modules into the extension_modules directory
    """
    if saltenv is None:
        saltenv = ["base"]

    if extmod_whitelist is None:
        extmod_whitelist = opts["extmod_whitelist"]
    elif isinstance(extmod_whitelist, str):
        extmod_whitelist = {form: extmod_whitelist.split(",")}
    elif not isinstance(extmod_whitelist, dict):
        log.error(
            "extmod_whitelist must be a string or dictionary: %s", extmod_whitelist
        )

    if extmod_blacklist is None:
        extmod_blacklist = opts["extmod_blacklist"]
    elif isinstance(extmod_blacklist, str):
        extmod_blacklist = {form: extmod_blacklist.split(",")}
    elif not isinstance(extmod_blacklist, dict):
        log.error(
            "extmod_blacklist must be a string or dictionary: %s", extmod_blacklist
        )

    if isinstance(saltenv, str):
        saltenv = saltenv.split(",")
    ret = []
    remote = set()
    source = salt.utils.url.create("_" + form)
    mod_dir = os.path.join(opts["extension_modules"], f"{form}")
    touched = False
    with salt.utils.files.set_umask(0o077):
        try:
            if not os.path.isdir(mod_dir):
                log.info("Creating module dir '%s'", mod_dir)
                try:
                    os.makedirs(mod_dir)
                except OSError:
                    log.error(
                        "Cannot create cache module directory %s. Check permissions.",
                        mod_dir,
                    )
            with salt.fileclient.get_file_client(opts) as fileclient:
                for sub_env in saltenv:
                    log.info("Syncing %s for environment '%s'", form, sub_env)
                    cache = []
                    log.info("Loading cache from %s, for %s", source, sub_env)
                    # Grab only the desired files (.py, .pyx, .so)
                    cache.extend(
                        fileclient.cache_dir(
                            source,
                            sub_env,
                            include_empty=False,
                            include_pat=r"E@\.(pyx?|so|zip)$",
                            exclude_pat=None,
                        )
                    )
                    local_cache_dir = os.path.join(
                        opts["cachedir"], "files", sub_env, f"_{form}"
                    )
                    log.debug("Local cache dir: '%s'", local_cache_dir)
                    for fn_ in cache:
                        relpath = os.path.relpath(fn_, local_cache_dir)
                        relname = os.path.splitext(relpath)[0].replace(os.sep, ".")
                        if (
                            extmod_whitelist
                            and form in extmod_whitelist
                            and relname not in extmod_whitelist[form]
                        ):
                            continue
                        if (
                            extmod_blacklist
                            and form in extmod_blacklist
                            and relname in extmod_blacklist[form]
                        ):
                            continue
                        remote.add(relpath)
                        dest = os.path.join(mod_dir, relpath)
                        log.info("Copying '%s' to '%s'", fn_, dest)
                        if os.path.isfile(dest):
                            # The file is present, if the sum differs replace it
                            hash_type = opts.get("hash_type", DEFAULT_HASH_TYPE)
                            src_digest = salt.utils.hashutils.get_hash(fn_, hash_type)
                            dst_digest = salt.utils.hashutils.get_hash(dest, hash_type)
                            if src_digest != dst_digest:
                                # The downloaded file differs, replace!
                                shutil.copyfile(fn_, dest)
                                ret.append(f"{form}.{relname}")
                        else:
                            dest_dir = os.path.dirname(dest)
                            if not os.path.isdir(dest_dir):
                                os.makedirs(dest_dir)
                            shutil.copyfile(fn_, dest)
                            ret.append(f"{form}.{relname}")

            touched = bool(ret)
            if opts["clean_dynamic_modules"] is True:
                current = set(_listdir_recursively(mod_dir))
                for fn_ in current - remote:
                    full = os.path.join(mod_dir, fn_)
                    if os.path.isfile(full):
                        touched = True
                        os.remove(full)
                # Cleanup empty dirs
                while True:
                    emptydirs = _list_emptydirs(mod_dir)
                    if not emptydirs:
                        break
                    for emptydir in emptydirs:
                        touched = True
                        shutil.rmtree(emptydir, ignore_errors=True)
        except Exception as exc:  # pylint: disable=broad-except
            log.error("Failed to sync %s module: %s", form, exc)

    return ret, touched



"""

We acquire the python information by running an interrogation script via subprocess trigger. This operation is not
cheap, especially not on Windows. To not have to pay this hefty cost every time we apply multiple levels of
caching.
"""  # noqa: D205

from __future__ import annotations

import hashlib
import logging
import os
import random
import sys
from collections import OrderedDict
from pathlib import Path
from shlex import quote
from string import ascii_lowercase, ascii_uppercase, digits
from subprocess import Popen

from virtualenv.app_data import AppDataDisabled
from virtualenv.discovery.py_info import PythonInfo
from virtualenv.util.subprocess import subprocess

_CACHE = OrderedDict()
_CACHE[Path(sys.executable)] = PythonInfo()
LOGGER = logging.getLogger(__name__)


def from_exe(cls, app_data, exe, env=None, raise_on_error=True, ignore_cache=False):  # noqa: FBT002, PLR0913
    env = os.environ if env is None else env
    result = _get_from_cache(cls, app_data, exe, env, ignore_cache=ignore_cache)
    if isinstance(result, Exception):
        if raise_on_error:
            raise result
        LOGGER.info("%s", result)
        result = None
    return result


def _get_from_cache(cls, app_data, exe, env, ignore_cache=True):  # noqa: FBT002
    # note here we cannot resolve symlinks, as the symlink may trigger different prefix information if there's a
    # pyenv.cfg somewhere alongside on python3.5+
    exe_path = Path(exe)
    if not ignore_cache and exe_path in _CACHE:  # check in the in-memory cache
        result = _CACHE[exe_path]
    else:  # otherwise go through the app data cache
        py_info = _get_via_file_cache(cls, app_data, exe_path, exe, env)
        result = _CACHE[exe_path] = py_info
    # independent if it was from the file or in-memory cache fix the original executable location
    if isinstance(result, PythonInfo):
        result.executable = exe
    return result


def _get_via_file_cache(cls, app_data, path, exe, env):
    path_text = str(path)
    try:
        path_modified = path.stat().st_mtime
    except OSError:
        path_modified = -1
    py_info_script = Path(os.path.abspath(__file__)).parent / "py_info.py"
    try:
        py_info_hash = hashlib.sha256(py_info_script.read_bytes()).hexdigest()
    except OSError:
        py_info_hash = None

    if app_data is None:
        app_data = AppDataDisabled()
    py_info, py_info_store = None, app_data.py_info(path)
    with py_info_store.locked():
        if py_info_store.exists():  # if exists and matches load
            data = py_info_store.read()
            of_path = data.get("path")
            of_st_mtime = data.get("st_mtime")
            of_content = data.get("content")
            of_hash = data.get("hash")
            if of_path == path_text and of_st_mtime == path_modified and of_hash == py_info_hash:
                py_info = cls._from_dict(of_content.copy())
                sys_exe = py_info.system_executable
                if sys_exe is not None and not os.path.exists(sys_exe):
                    py_info_store.remove()
                    py_info = None
            else:
                py_info_store.remove()
        if py_info is None:  # if not loaded run and save
            failure, py_info = _run_subprocess(cls, exe, app_data, env)
            if failure is None:
                data = {
                    "st_mtime": path_modified,
                    "path": path_text,
                    "content": py_info._to_dict(),  # noqa: SLF001
                    "hash": py_info_hash,
                }
                py_info_store.write(data)
            else:
                py_info = failure
    return py_info


COOKIE_LENGTH: int = 32


def gen_cookie():
    return "".join(
        random.choice(f"{ascii_lowercase}{ascii_uppercase}{digits}")  # noqa: S311
        for _ in range(COOKIE_LENGTH)
    )


def _run_subprocess(cls, exe, app_data, env):
    py_info_script = Path(os.path.abspath(__file__)).parent / "py_info.py"
    # Cookies allow to split the serialized stdout output generated by the script collecting the info from the output
    # generated by something else. The right way to deal with it is to create an anonymous pipe and pass its descriptor
    # to the child and output to it. But AFAIK all of them are either not cross-platform or too big to implement and are
    # not in the stdlib. So the easiest and the shortest way I could mind is just using the cookies.
    # We generate pseudorandom cookies because it easy to implement and avoids breakage from outputting modules source
    # code, i.e. by debug output libraries. We reverse the cookies to avoid breakages resulting from variable values
    # appearing in debug output.

    start_cookie = gen_cookie()
    end_cookie = gen_cookie()
    with app_data.ensure_extracted(py_info_script) as py_info_script:
        cmd = [exe, str(py_info_script), start_cookie, end_cookie]
        # prevent sys.prefix from leaking into the child process - see https://bugs.python.org/issue22490
        env = env.copy()
        env.pop("__PYVENV_LAUNCHER__", None)
        # Force UTF-8 mode to handle output from Microsoft Store Python stub on Windows
        # See https://github.com/pypa/virtualenv/issues/2812
        env["PYTHONUTF8"] = "1"
        LOGGER.debug("get interpreter info via cmd: %s", LogCmd(cmd))
        try:
            process = Popen(
                cmd,
                universal_newlines=True,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
                env=env,
                encoding="utf-8",
                errors="backslashreplace",
            )
            out, err = process.communicate()
            code = process.returncode
        except OSError as os_error:
            out, err, code = "", os_error.strerror, os_error.errno
    result, failure = None, None
    if code == 0:
        out_starts = out.find(start_cookie[::-1])

        if out_starts > -1:
            pre_cookie = out[:out_starts]

            if pre_cookie:
                sys.stdout.write(pre_cookie)

            out = out[out_starts + COOKIE_LENGTH :]

        out_ends = out.find(end_cookie[::-1])

        if out_ends > -1:
            post_cookie = out[out_ends + COOKIE_LENGTH :]

            if post_cookie:
                sys.stdout.write(post_cookie)

            out = out[:out_ends]

        result = cls._from_json(out)
        result.executable = exe  # keep original executable as this may contain initialization code
    else:
        msg = f"{exe} with code {code}{f' out: {out!r}' if out else ''}{f' err: {err!r}' if err else ''}"
        failure = RuntimeError(f"failed to query {msg}")
    return failure, result


class LogCmd:
    def __init__(self, cmd, env=None) -> None:
        self.cmd = cmd
        self.env = env

    def __repr__(self) -> str:
        cmd_repr = " ".join(quote(str(c)) for c in self.cmd)
        if self.env is not None:
            cmd_repr = f"{cmd_repr} env of {self.env!r}"
        return cmd_repr


def clear(app_data):
    app_data.py_info_clear()
    _CACHE.clear()


___all___ = [
    "from_exe",
    "clear",
    "LogCmd",
]



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging

from opentelemetry import metrics
from opentelemetry.sdk._configuration._common import (
    _map_compression,
    _parse_headers,
    load_entry_point,
)
from opentelemetry.sdk._configuration._exceptions import ConfigurationError
from opentelemetry.sdk._configuration.models import (
    Aggregation as AggregationConfig,
)
from opentelemetry.sdk._configuration.models import (
    ConsoleMetricExporter as ConsoleMetricExporterConfig,
)
from opentelemetry.sdk._configuration.models import (
    ExemplarFilter as ExemplarFilterConfig,
)
from opentelemetry.sdk._configuration.models import (
    ExporterDefaultHistogramAggregation,
    ExporterTemporalityPreference,
    InstrumentType,
)
from opentelemetry.sdk._configuration.models import (
    MeterProvider as MeterProviderConfig,
)
from opentelemetry.sdk._configuration.models import (
    MetricReader as MetricReaderConfig,
)
from opentelemetry.sdk._configuration.models import (
    OtlpGrpcMetricExporter as OtlpGrpcMetricExporterConfig,
)
from opentelemetry.sdk._configuration.models import (
    OtlpHttpMetricExporter as OtlpHttpMetricExporterConfig,
)
from opentelemetry.sdk._configuration.models import (
    PeriodicMetricReader as PeriodicMetricReaderConfig,
)
from opentelemetry.sdk._configuration.models import (
    PushMetricExporter as PushMetricExporterConfig,
)
from opentelemetry.sdk._configuration.models import (
    View as ViewConfig,
)
from opentelemetry.sdk.metrics import (
    AlwaysOffExemplarFilter,
    AlwaysOnExemplarFilter,
    Counter,
    Histogram,
    MeterProvider,
    ObservableCounter,
    ObservableGauge,
    ObservableUpDownCounter,
    TraceBasedExemplarFilter,
    UpDownCounter,
    _Gauge,
)
from opentelemetry.sdk.metrics.export import (
    AggregationTemporality,
    ConsoleMetricExporter,
    MetricExporter,
    MetricReader,
    PeriodicExportingMetricReader,
)
from opentelemetry.sdk.metrics.view import (
    Aggregation,
    DefaultAggregation,
    DropAggregation,
    ExplicitBucketHistogramAggregation,
    ExponentialBucketHistogramAggregation,
    LastValueAggregation,
    SumAggregation,
    View,
)
from opentelemetry.sdk.resources import Resource

_logger = logging.getLogger(__name__)


# Default interval/timeout per OTel spec (milliseconds).
_DEFAULT_EXPORT_INTERVAL_MILLIS = 60000
_DEFAULT_EXPORT_TIMEOUT_MILLIS = 30000

# Instrument type → SDK instrument class mapping (for View selectors).
_INSTRUMENT_TYPE_MAP: dict[InstrumentType, type] = {
    InstrumentType.counter: Counter,
    InstrumentType.up_down_counter: UpDownCounter,
    InstrumentType.histogram: Histogram,
    InstrumentType.gauge: _Gauge,
    InstrumentType.observable_counter: ObservableCounter,
    InstrumentType.observable_gauge: ObservableGauge,
    InstrumentType.observable_up_down_counter: ObservableUpDownCounter,
}


def _map_temporality(
    pref: ExporterTemporalityPreference | None,
) -> dict[type, AggregationTemporality]:
    """Map a temporality preference to an explicit preferred_temporality dict.

    Always returns an explicit dict to suppress OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE.
    Default (None or cumulative) → all instruments CUMULATIVE.
    """
    if pref is None or pref == ExporterTemporalityPreference.cumulative:
        return {
            Counter: AggregationTemporality.CUMULATIVE,
            UpDownCounter: AggregationTemporality.CUMULATIVE,
            Histogram: AggregationTemporality.CUMULATIVE,
            ObservableCounter: AggregationTemporality.CUMULATIVE,
            ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
            ObservableGauge: AggregationTemporality.CUMULATIVE,
        }
    if pref == ExporterTemporalityPreference.delta:
        return {
            Counter: AggregationTemporality.DELTA,
            UpDownCounter: AggregationTemporality.CUMULATIVE,
            Histogram: AggregationTemporality.DELTA,
            ObservableCounter: AggregationTemporality.DELTA,
            ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
            ObservableGauge: AggregationTemporality.CUMULATIVE,
        }
    if pref == ExporterTemporalityPreference.low_memory:
        return {
            Counter: AggregationTemporality.DELTA,
            UpDownCounter: AggregationTemporality.CUMULATIVE,
            Histogram: AggregationTemporality.DELTA,
            ObservableCounter: AggregationTemporality.CUMULATIVE,
            ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
            ObservableGauge: AggregationTemporality.CUMULATIVE,
        }
    raise ConfigurationError(
        f"Unsupported temporality preference '{pref}'. "
        "Supported values: cumulative, delta, low_memory."
    )


def _map_histogram_aggregation(
    pref: ExporterDefaultHistogramAggregation | None,
) -> dict[type, Aggregation]:
    """Map a histogram aggregation preference to an explicit preferred_aggregation dict.

    Always returns an explicit dict to suppress
    OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION.
    Default (None or explicit_bucket_histogram) → ExplicitBucketHistogramAggregation.
    """
    if (
        pref is None
        or pref
        == ExporterDefaultHistogramAggregation.explicit_bucket_histogram
    ):
        return {Histogram: ExplicitBucketHistogramAggregation()}
    if (
        pref
        == ExporterDefaultHistogramAggregation.base2_exponential_bucket_histogram
    ):
        return {Histogram: ExponentialBucketHistogramAggregation()}
    raise ConfigurationError(
        f"Unsupported default histogram aggregation '{pref}'. "
        "Supported values: explicit_bucket_histogram, base2_exponential_bucket_histogram."
    )


def _create_aggregation(config: AggregationConfig) -> Aggregation:
    """Create an SDK Aggregation from config, passing through detail parameters."""
    if config.default is not None:
        return DefaultAggregation()
    if config.drop is not None:
        return DropAggregation()
    if config.explicit_bucket_histogram is not None:
        return ExplicitBucketHistogramAggregation(
            boundaries=config.explicit_bucket_histogram.boundaries,
            record_min_max=(
                config.explicit_bucket_histogram.record_min_max
                if config.explicit_bucket_histogram.record_min_max is not None
                else True
            ),
        )
    if config.base2_exponential_bucket_histogram is not None:
        kwargs = {}
        if config.base2_exponential_bucket_histogram.max_size is not None:
            kwargs["max_size"] = (
                config.base2_exponential_bucket_histogram.max_size
            )
        if config.base2_exponential_bucket_histogram.max_scale is not None:
            kwargs["max_scale"] = (
                config.base2_exponential_bucket_histogram.max_scale
            )
        return ExponentialBucketHistogramAggregation(**kwargs)
    if config.last_value is not None:
        return LastValueAggregation()
    if config.sum is not None:
        return SumAggregation()
    raise ConfigurationError(
        f"Unknown or unsupported aggregation type in config: {config!r}. "
        "Supported types: default, drop, explicit_bucket_histogram, "
        "base2_exponential_bucket_histogram, last_value, sum."
    )


def _create_view(config: ViewConfig) -> View:
    """Create an SDK View from config."""
    selector = config.selector
    stream = config.stream

    instrument_type = None
    if selector.instrument_type is not None:
        instrument_type = _INSTRUMENT_TYPE_MAP.get(selector.instrument_type)
        if instrument_type is None:
            raise ConfigurationError(
                f"Unknown instrument type: {selector.instrument_type!r}"
            )

    attribute_keys: set[str] | None = None
    if stream.attribute_keys is not None:
        if stream.attribute_keys.excluded:
            _logger.warning(
                "attribute_keys.excluded is not supported by the Python SDK View; "
                "the exclusion list will be ignored."
            )
        if stream.attribute_keys.included is not None:
            attribute_keys = set(stream.attribute_keys.included)

    aggregation = None
    if stream.aggregation is not None:
        aggregation = _create_aggregation(stream.aggregation)

    return View(
        instrument_type=instrument_type,
        instrument_name=selector.instrument_name,
        meter_name=selector.meter_name,
        meter_version=selector.meter_version,
        meter_schema_url=selector.meter_schema_url,
        instrument_unit=selector.unit,
        name=stream.name,
        description=stream.description,
        attribute_keys=attribute_keys,
        aggregation=aggregation,
    )


def _create_console_metric_exporter(
    config: ConsoleMetricExporterConfig,
) -> MetricExporter:
    """Create a ConsoleMetricExporter from config."""
    preferred_temporality = _map_temporality(config.temporality_preference)
    preferred_aggregation = _map_histogram_aggregation(
        config.default_histogram_aggregation
    )
    return ConsoleMetricExporter(
        preferred_temporality=preferred_temporality,
        preferred_aggregation=preferred_aggregation,
    )


def _create_otlp_http_metric_exporter(
    config: OtlpHttpMetricExporterConfig,
) -> MetricExporter:
    """Create an OTLP HTTP metric exporter from config."""
    try:
        # pylint: disable=import-outside-toplevel,no-name-in-module
        from opentelemetry.exporter.otlp.proto.http import (  # type: ignore[import-untyped]  # noqa: PLC0415
            Compression,
        )
        from opentelemetry.exporter.otlp.proto.http.metric_exporter import (  # type: ignore[import-untyped]  # noqa: PLC0415
            OTLPMetricExporter,
        )
    except ImportError as exc:
        raise ConfigurationError(
            "otlp_http metric exporter requires 'opentelemetry-exporter-otlp-proto-http'. "
            "Install it with: pip install opentelemetry-exporter-otlp-proto-http"
        ) from exc

    compression = _map_compression(
        config.compression, Compression, allow_deflate=True
    )
    headers = _parse_headers(config.headers, config.headers_list)
    timeout = (config.timeout / 1000.0) if config.timeout is not None else None
    preferred_temporality = _map_temporality(config.temporality_preference)
    preferred_aggregation = _map_histogram_aggregation(
        config.default_histogram_aggregation
    )

    return OTLPMetricExporter(  # type: ignore[return-value]
        endpoint=config.endpoint,
        headers=headers,
        timeout=timeout,
        compression=compression,  # type: ignore[arg-type]
        preferred_temporality=preferred_temporality,
        preferred_aggregation=preferred_aggregation,
    )


def _create_otlp_grpc_metric_exporter(
    config: OtlpGrpcMetricExporterConfig,
) -> MetricExporter:
    """Create an OTLP gRPC metric exporter from config."""
    try:
        # pylint: disable=import-outside-toplevel,no-name-in-module
        import grpc  # type: ignore[import-untyped]  # noqa: PLC0415

        from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (  # type: ignore[import-untyped]  # noqa: PLC0415
            OTLPMetricExporter,
        )
    except ImportError as exc:
        raise ConfigurationError(
            "otlp_grpc metric exporter requires 'opentelemetry-exporter-otlp-proto-grpc'. "
            "Install it with: pip install opentelemetry-exporter-otlp-proto-grpc"
        ) from exc

    compression = _map_compression(config.compression, grpc.Compression)
    headers = _parse_headers(config.headers, config.headers_list)
    timeout = (config.timeout / 1000.0) if config.timeout is not None else None
    preferred_temporality = _map_temporality(config.temporality_preference)
    preferred_aggregation = _map_histogram_aggregation(
        config.default_histogram_aggregation
    )

    return OTLPMetricExporter(  # type: ignore[return-value]
        endpoint=config.endpoint,
        headers=headers,
        timeout=timeout,
        compression=compression,  # type: ignore[arg-type]
        preferred_temporality=preferred_temporality,
        preferred_aggregation=preferred_aggregation,
    )


_METRIC_EXPORTER_REGISTRY: dict = {
    "otlp_http": _create_otlp_http_metric_exporter,
    "otlp_grpc": _create_otlp_grpc_metric_exporter,
    "console": _create_console_metric_exporter,
}


def _create_push_metric_exporter(
    config: PushMetricExporterConfig,
) -> MetricExporter:
    """Create a push metric exporter from config.

    Known exporter types are checked via typed fields on the PushMetricExporter
    dataclass. Unknown exporter names captured in additional_properties
    by the @_additional_properties decorator are loaded via the
    ``opentelemetry_metrics_exporter`` entry point group.
    """
    if config.otlp_file_development is not None:
        raise ConfigurationError(
            "otlp_file_development metric exporter is experimental "
            "and not yet supported."
        )
    for name, factory in _METRIC_EXPORTER_REGISTRY.items():
        value = getattr(config, name, None)
        if value is not None:
            return factory(value)
    if config.additional_properties:
        name, plugin_config = next(iter(config.additional_properties.items()))
        return load_entry_point("opentelemetry_metrics_exporter", name)(
            **(plugin_config or {})
        )
    raise ConfigurationError(
        "No exporter type specified in push metric exporter config. "
        "Supported types: console, otlp_http, otlp_grpc."
    )


def _create_periodic_metric_reader(
    config: PeriodicMetricReaderConfig,
) -> PeriodicExportingMetricReader:
    """Create a PeriodicExportingMetricReader from config.

    Passes explicit interval/timeout defaults to suppress env var reading.
    """
    exporter = _create_push_metric_exporter(config.exporter)
    interval = (
        config.interval
        if config.interval is not None
        else _DEFAULT_EXPORT_INTERVAL_MILLIS
    )
    timeout = (
        config.timeout
        if config.timeout is not None
        else _DEFAULT_EXPORT_TIMEOUT_MILLIS
    )
    return PeriodicExportingMetricReader(
        exporter=exporter,
        export_interval_millis=float(interval),
        export_timeout_millis=float(timeout),
    )


def _create_metric_reader(config: MetricReaderConfig) -> MetricReader:
    """Create a MetricReader from config."""
    if config.periodic is not None:
        return _create_periodic_metric_reader(config.periodic)
    if config.pull is not None:
        raise ConfigurationError(
            "Pull metric readers (e.g. Prometheus) are experimental and not yet supported "
            "by declarative config. Use the SDK API directly to configure pull readers."
        )
    raise ConfigurationError(
        "No reader type specified in metric reader config. "
        "Supported types: periodic."
    )


def _create_exemplar_filter(
    value: ExemplarFilterConfig,
) -> object:
    """Create an SDK exemplar filter from config enum value."""
    if value == ExemplarFilterConfig.always_on:
        return AlwaysOnExemplarFilter()
    if value == ExemplarFilterConfig.always_off:
        return AlwaysOffExemplarFilter()
    if value == ExemplarFilterConfig.trace_based:
        return TraceBasedExemplarFilter()
    raise ConfigurationError(
        f"Unknown exemplar filter value: {value!r}. "
        "Supported values: always_on, always_off, trace_based."
    )


def create_meter_provider(
    config: MeterProviderConfig | None,
    resource: Resource | None = None,
) -> MeterProvider:
    """Create an SDK MeterProvider from declarative config.

    Does NOT read OTEL_METRIC_EXPORT_INTERVAL, OTEL_METRICS_EXEMPLAR_FILTER,
    or any other env vars for values explicitly controlled by the config.
    Absent config values use OTel spec defaults, matching Java SDK behavior.

    Args:
        config: MeterProvider config from the parsed config file, or None.
        resource: Resource to attach to the provider.

    Returns:
        A configured MeterProvider.
    """
    # Always pass an explicit exemplar filter to suppress env var reading.
    # Spec default is trace_based.
    exemplar_filter: object = TraceBasedExemplarFilter()
    if config is not None and config.exemplar_filter is not None:
        exemplar_filter = _create_exemplar_filter(config.exemplar_filter)

    readers: list[MetricReader] = []
    views: list[View] = []

    if config is not None:
        for reader_config in config.readers:
            readers.append(_create_metric_reader(reader_config))
        if config.views:
            for view_config in config.views:
                views.append(_create_view(view_config))

    return MeterProvider(
        resource=resource,
        metric_readers=readers,
        exemplar_filter=exemplar_filter,  # type: ignore[arg-type]
        views=views,
    )


def configure_meter_provider(
    config: MeterProviderConfig | None,
    resource: Resource | None = None,
) -> None:
    """Configure the global MeterProvider from declarative config.

    When config is None (meter_provider section absent from config file),
    the global is not set — matching Java/JS SDK behavior.

    Args:
        config: MeterProvider config from the parsed config file, or None.
        resource: Resource to attach to the provider.
    """
    if config is None:
        return
    metrics.set_meter_provider(create_meter_provider(config, resource))



# Copyright (C) 2010, 2011 Sebastian Thiel (byronimo@gmail.com) and contributors
#
# This module is part of GitDB and is released under
# the New BSD License: https://opensource.org/license/bsd-3-clause/
"""Module containing information about types known to the database"""

str_blob_type = b'blob'
str_commit_type = b'commit'
str_tree_type = b'tree'
str_tag_type = b'tag'



# pylint: disable=import-error
try:
    import pyrax

    from salt.utils.openstack.pyrax.authenticate import Authenticate
    from salt.utils.openstack.pyrax.queues import RackspaceQueues

    __all__ = ["Authenticate", "RackspaceQueues"]

    HAS_PYRAX = True
except ImportError as err:
    HAS_PYRAX = False
# pylint: enable=import-error



import os

import salt.cli.caller
import salt.defaults.exitcodes
import salt.utils.parsers
import salt.utils.tracing
import salt.utils.verify
from salt.config import _expand_glob_path, prepend_root_dir


class SaltCall(salt.utils.parsers.SaltCallOptionParser):
    """
    Used to locally execute a salt command
    """

    def run(self):
        """
        Execute the salt call!
        """
        self.parse_args()

        if self.options.file_root:
            file_roots = []
            for file_root in self.options.file_root:
                # check if the argument is pointing to a file on disk
                file_roots.append(os.path.abspath(file_root))
            self.config["file_roots"] = {"base": _expand_glob_path(file_roots)}

        if self.options.pillar_root:
            pillar_roots = []
            for pillar_root in self.options.pillar_root:
                # check if the argument is pointing to a file on disk
                pillar_roots.append(os.path.abspath(pillar_root))
            self.config["pillar_roots"] = {"base": _expand_glob_path(pillar_roots)}

        if self.options.states_dir:
            states_dirs = []
            for states_dir in self.options.states_dir:
                # check if the argument is pointing to a file on disk
                states_dirs.append(os.path.abspath(states_dir))
            self.config["states_dirs"] = states_dirs

        if self.options.local:
            self.config["file_client"] = "local"
        if self.options.master:
            self.config["master"] = self.options.master

        if self.options.cachedir and self.config.get(
            "extension_modules"
        ) == os.path.join(self.config.get("__cachedir"), "extmods"):
            # Override `extension_modules`, but only in case if it was autogenerated
            cache_dir = os.path.abspath(self.options.cachedir)
            self.config["cachedir"] = cache_dir
            self.config["extension_modules"] = os.path.join(cache_dir, "extmods")
            prepend_root_dir(self.config, ["cachedir", "extension_modules"])

        if self.config["verify_env"]:
            # When --priv is used, MergeConfigMixIn has already overwritten
            # config["user"] with the --priv value during parse_args().  We need
            # the user that is actually *configured* in the config files so that
            # verify_env chowns directories to the right owner (e.g. 'salt') rather
            # than the temporary execution-time override (e.g. 'root').
            if self.options.user:
                import salt.config as _salt_config

                _raw_opts = _salt_config.minion_config(
                    self.get_config_file_path(), cache_minion_id=False
                )
                _verify_user = _raw_opts.get("user", "root")
            else:
                _verify_user = self.config["user"]

            salt.utils.verify.verify_env(
                [
                    self.config["pki_dir"],
                    self.config["cachedir"],
                    self.config["extension_modules"],
                ],
                _verify_user,
                permissive=self.config["permissive_pki_access"],
                pki_dir=self.config["pki_dir"],
            )

        # config["user"] is already set to the --priv value by MergeConfigMixIn.
        # The explicit assignment below is kept for clarity and backward compatibility
        # in case MergeConfigMixIn behaviour ever changes.
        if self.options.user:
            self.config["user"] = self.options.user

        # Validate the execution user exists
        if self.config["user"] != salt.utils.user.get_user():
            if not salt.utils.verify.check_user(self.config["user"]):
                self.exit(salt.defaults.exitcodes.EX_NOUSER)

        caller = salt.cli.caller.Caller.factory(self.config)

        if self.options.doc:
            caller.print_docs()
            self.exit(salt.defaults.exitcodes.EX_OK)

        if self.options.grains_run:
            caller.print_grains()
            self.exit(salt.defaults.exitcodes.EX_OK)

        salt.utils.tracing.configure({**self.config, "__role": "minion"})
        env_carrier = {
            k: v
            for k, v in (
                ("traceparent", os.environ.get("TRACEPARENT", "")),
                ("tracestate", os.environ.get("TRACESTATE", "")),
            )
            if v
        }
        trace_ctx = salt.utils.tracing.extract(env_carrier)
        with salt.utils.tracing.start_span(
            (
                f"salt-call.{self.config.get('fun', '')}"
                if self.config.get("fun")
                else "salt-call"
            ),
            attributes={"salt.fun": self.config.get("fun", "")},
            context=trace_ctx,
        ):
            try:
                caller.run()
            finally:
                salt.utils.tracing.shutdown()



# Copyright 2014-2016 Nathan West
#
# This file is part of autocommand.
#
# autocommand is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# autocommand is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with autocommand.  If not, see <http://www.gnu.org/licenses/>.

# flake8 flags all these imports as unused, hence the NOQAs everywhere.

from .automain import automain  # NOQA
from .autoparse import autoparse, smart_open  # NOQA
from .autocommand import autocommand  # NOQA

try:
    from .autoasync import autoasync  # NOQA
except ImportError:  # pragma: no cover
    pass



import math
import json

from jmespath import exceptions
from jmespath.compat import string_type as STRING_TYPE
from jmespath.compat import get_methods


# python types -> jmespath types
TYPES_MAP = {
    'bool': 'boolean',
    'list': 'array',
    'dict': 'object',
    'NoneType': 'null',
    'unicode': 'string',
    'str': 'string',
    'float': 'number',
    'int': 'number',
    'long': 'number',
    'OrderedDict': 'object',
    '_Projection': 'array',
    '_Expression': 'expref',
}


# jmespath types -> python types
REVERSE_TYPES_MAP = {
    'boolean': ('bool',),
    'array': ('list', '_Projection'),
    'object': ('dict', 'OrderedDict',),
    'null': ('NoneType',),
    'string': ('unicode', 'str'),
    'number': ('float', 'int', 'long'),
    'expref': ('_Expression',),
}


def signature(*arguments):
    def _record_signature(func):
        func.signature = arguments
        return func
    return _record_signature


class FunctionRegistry(type):
    def __init__(cls, name, bases, attrs):
        cls._populate_function_table()
        super(FunctionRegistry, cls).__init__(name, bases, attrs)

    def _populate_function_table(cls):
        function_table = {}
        # Any method with a @signature decorator that also
        # starts with "_func_" is registered as a function.
        # _func_max_by -> max_by function.
        for name, method in get_methods(cls):
            if not name.startswith('_func_'):
                continue
            signature = getattr(method, 'signature', None)
            if signature is not None:
                function_table[name[6:]] = {
                    'function': method,
                    'signature': signature,
                }
        cls.FUNCTION_TABLE = function_table


class Functions(metaclass=FunctionRegistry):

    FUNCTION_TABLE = {
    }

    def call_function(self, function_name, resolved_args):
        try:
            spec = self.FUNCTION_TABLE[function_name]
        except KeyError:
            raise exceptions.UnknownFunctionError(
                "Unknown function: %s()" % function_name)
        function = spec['function']
        signature = spec['signature']
        self._validate_arguments(resolved_args, signature, function_name)
        return function(self, *resolved_args)

    def _validate_arguments(self, args, signature, function_name):
        if signature and signature[-1].get('variadic'):
            if len(args) < len(signature):
                raise exceptions.VariadictArityError(
                    len(signature), len(args), function_name)
        elif len(args) != len(signature):
            raise exceptions.ArityError(
                len(signature), len(args), function_name)
        return self._type_check(args, signature, function_name)

    def _type_check(self, actual, signature, function_name):
        for i in range(len(signature)):
            allowed_types = signature[i]['types']
            if allowed_types:
                self._type_check_single(actual[i], allowed_types,
                                        function_name)

    def _type_check_single(self, current, types, function_name):
        # Type checking involves checking the top level type,
        # and in the case of arrays, potentially checking the types
        # of each element.
        allowed_types, allowed_subtypes = self._get_allowed_pytypes(types)
        # We're not using isinstance() on purpose.
        # The type model for jmespath does not map
        # 1-1 with python types (booleans are considered
        # integers in python for example).
        actual_typename = type(current).__name__
        if actual_typename not in allowed_types:
            raise exceptions.JMESPathTypeError(
                function_name, current,
                self._convert_to_jmespath_type(actual_typename), types)
        # If we're dealing with a list type, we can have
        # additional restrictions on the type of the list
        # elements (for example a function can require a
        # list of numbers or a list of strings).
        # Arrays are the only types that can have subtypes.
        if allowed_subtypes:
            self._subtype_check(current, allowed_subtypes,
                                types, function_name)

    def _get_allowed_pytypes(self, types):
        allowed_types = []
        allowed_subtypes = []
        for t in types:
            type_ = t.split('-', 1)
            if len(type_) == 2:
                type_, subtype = type_
                allowed_subtypes.append(REVERSE_TYPES_MAP[subtype])
            else:
                type_ = type_[0]
            allowed_types.extend(REVERSE_TYPES_MAP[type_])
        return allowed_types, allowed_subtypes

    def _subtype_check(self, current, allowed_subtypes, types, function_name):
        if len(allowed_subtypes) == 1:
            # The easy case, we know up front what type
            # we need to validate.
            allowed_subtypes = allowed_subtypes[0]
            for element in current:
                actual_typename = type(element).__name__
                if actual_typename not in allowed_subtypes:
                    raise exceptions.JMESPathTypeError(
                        function_name, element, actual_typename, types)
        elif len(allowed_subtypes) > 1 and current:
            # Dynamic type validation.  Based on the first
            # type we see, we validate that the remaining types
            # match.
            first = type(current[0]).__name__
            for subtypes in allowed_subtypes:
                if first in subtypes:
                    allowed = subtypes
                    break
            else:
                raise exceptions.JMESPathTypeError(
                    function_name, current[0], first, types)
            for element in current:
                actual_typename = type(element).__name__
                if actual_typename not in allowed:
                    raise exceptions.JMESPathTypeError(
                        function_name, element, actual_typename, types)

    @signature({'types': ['number']})
    def _func_abs(self, arg):
        return abs(arg)

    @signature({'types': ['array-number']})
    def _func_avg(self, arg):
        if arg:
            return sum(arg) / len(arg)
        else:
            return None

    @signature({'types': [], 'variadic': True})
    def _func_not_null(self, *arguments):
        for argument in arguments:
            if argument is not None:
                return argument

    @signature({'types': []})
    def _func_to_array(self, arg):
        if isinstance(arg, list):
            return arg
        else:
            return [arg]

    @signature({'types': []})
    def _func_to_string(self, arg):
        if isinstance(arg, STRING_TYPE):
            return arg
        else:
            return json.dumps(arg, separators=(',', ':'),
                              default=str)

    @signature({'types': []})
    def _func_to_number(self, arg):
        if isinstance(arg, (list, dict, bool)):
            return None
        elif arg is None:
            return None
        elif isinstance(arg, (int, float)):
            return arg
        else:
            try:
                return int(arg)
            except ValueError:
                try:
                    return float(arg)
                except ValueError:
                    return None

    @signature({'types': ['array', 'string']}, {'types': []})
    def _func_contains(self, subject, search):
        return search in subject

    @signature({'types': ['string', 'array', 'object']})
    def _func_length(self, arg):
        return len(arg)

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_ends_with(self, search, suffix):
        return search.endswith(suffix)

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_starts_with(self, search, suffix):
        return search.startswith(suffix)

    @signature({'types': ['array', 'string']})
    def _func_reverse(self, arg):
        if isinstance(arg, STRING_TYPE):
            return arg[::-1]
        else:
            return list(reversed(arg))

    @signature({"types": ['number']})
    def _func_ceil(self, arg):
        return math.ceil(arg)

    @signature({"types": ['number']})
    def _func_floor(self, arg):
        return math.floor(arg)

    @signature({"types": ['string']}, {"types": ['array-string']})
    def _func_join(self, separator, array):
        return separator.join(array)

    @signature({'types': ['expref']}, {'types': ['array']})
    def _func_map(self, expref, arg):
        result = []
        for element in arg:
            result.append(expref.visit(expref.expression, element))
        return result

    @signature({"types": ['array-number', 'array-string']})
    def _func_max(self, arg):
        if arg:
            return max(arg)
        else:
            return None

    @signature({"types": ["object"], "variadic": True})
    def _func_merge(self, *arguments):
        merged = {}
        for arg in arguments:
            merged.update(arg)
        return merged

    @signature({"types": ['array-number', 'array-string']})
    def _func_min(self, arg):
        if arg:
            return min(arg)
        else:
            return None

    @signature({"types": ['array-string', 'array-number']})
    def _func_sort(self, arg):
        return list(sorted(arg))

    @signature({"types": ['array-number']})
    def _func_sum(self, arg):
        return sum(arg)

    @signature({"types": ['object']})
    def _func_keys(self, arg):
        # To be consistent with .values()
        # should we also return the indices of a list?
        return list(arg.keys())

    @signature({"types": ['object']})
    def _func_values(self, arg):
        return list(arg.values())

    @signature({'types': []})
    def _func_type(self, arg):
        if isinstance(arg, STRING_TYPE):
            return "string"
        elif isinstance(arg, bool):
            return "boolean"
        elif isinstance(arg, list):
            return "array"
        elif isinstance(arg, dict):
            return "object"
        elif isinstance(arg, (float, int)):
            return "number"
        elif arg is None:
            return "null"

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_sort_by(self, array, expref):
        if not array:
            return array
        # sort_by allows for the expref to be either a number of
        # a string, so we have some special logic to handle this.
        # We evaluate the first array element and verify that it's
        # either a string of a number.  We then create a key function
        # that validates that type, which requires that remaining array
        # elements resolve to the same type as the first element.
        required_type = self._convert_to_jmespath_type(
            type(expref.visit(expref.expression, array[0])).__name__)
        if required_type not in ['number', 'string']:
            raise exceptions.JMESPathTypeError(
                'sort_by', array[0], required_type, ['string', 'number'])
        keyfunc = self._create_key_func(expref,
                                        [required_type],
                                        'sort_by')
        return list(sorted(array, key=keyfunc))

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_min_by(self, array, expref):
        keyfunc = self._create_key_func(expref,
                                        ['number', 'string'],
                                        'min_by')
        if array:
            return min(array, key=keyfunc)
        else:
            return None

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_max_by(self, array, expref):
        keyfunc = self._create_key_func(expref,
                                        ['number', 'string'],
                                        'max_by')
        if array:
            return max(array, key=keyfunc)
        else:
            return None

    def _create_key_func(self, expref, allowed_types, function_name):
        def keyfunc(x):
            result = expref.visit(expref.expression, x)
            actual_typename = type(result).__name__
            jmespath_type = self._convert_to_jmespath_type(actual_typename)
            # allowed_types is in term of jmespath types, not python types.
            if jmespath_type not in allowed_types:
                raise exceptions.JMESPathTypeError(
                    function_name, result, jmespath_type, allowed_types)
            return result
        return keyfunc

    def _convert_to_jmespath_type(self, pyobject):
        return TYPES_MAP.get(pyobject, 'unknown')



# This module is part of GitPython and is released under the
# 3-Clause BSD License: https://opensource.org/license/bsd-3-clause/

__all__ = [
    "sm_section",
    "sm_name",
    "mkhead",
    "find_first_remote_branch",
    "SubmoduleConfigParser",
]

from io import BytesIO
import weakref

import git
from git.config import GitConfigParser
from git.exc import InvalidGitRepositoryError

# typing -----------------------------------------------------------------------

from typing import Any, Sequence, TYPE_CHECKING, Union

from git.types import PathLike

if TYPE_CHECKING:
    from weakref import ReferenceType

    from git.refs import Head, RemoteReference
    from git.remote import Remote
    from git.repo import Repo

    from .base import Submodule

# { Utilities


def sm_section(name: str) -> str:
    """:return: Section title used in ``.gitmodules`` configuration file"""
    return f'submodule "{name}"'


def sm_name(section: str) -> str:
    """:return: Name of the submodule as parsed from the section name"""
    section = section.strip()
    return section[11:-1]


def mkhead(repo: "Repo", path: PathLike) -> "Head":
    """:return: New branch/head instance"""
    return git.Head(repo, git.Head.to_full_path(path))


def find_first_remote_branch(remotes: Sequence["Remote"], branch_name: str) -> "RemoteReference":
    """Find the remote branch matching the name of the given branch or raise
    :exc:`~git.exc.InvalidGitRepositoryError`."""
    for remote in remotes:
        try:
            return remote.refs[branch_name]
        except IndexError:
            continue
        # END exception handling
    # END for remote
    raise InvalidGitRepositoryError("Didn't find remote branch '%r' in any of the given remotes" % branch_name)


# } END utilities

# { Classes


class SubmoduleConfigParser(GitConfigParser):
    """Catches calls to :meth:`~git.config.GitConfigParser.write`, and updates the
    ``.gitmodules`` blob in the index with the new data, if we have written into a
    stream.

    Otherwise it would add the local file to the index to make it correspond with the
    working tree. Additionally, the cache must be cleared.

    Please note that no mutating method will work in bare mode.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._smref: Union["ReferenceType[Submodule]", None] = None
        self._index = None
        self._auto_write = True
        super().__init__(*args, **kwargs)

    # { Interface
    def set_submodule(self, submodule: "Submodule") -> None:
        """Set this instance's submodule. It must be called before the first write
        operation begins."""
        self._smref = weakref.ref(submodule)

    def flush_to_index(self) -> None:
        """Flush changes in our configuration file to the index."""
        assert self._smref is not None
        # Should always have a file here.
        assert not isinstance(self._file_or_files, BytesIO)

        sm = self._smref()
        if sm is not None:
            index = self._index
            if index is None:
                index = sm.repo.index
            # END handle index
            index.add([sm.k_modules_file], write=self._auto_write)
            sm._clear_cache()
        # END handle weakref

    # } END interface

    # { Overridden Methods
    def write(self) -> None:  # type: ignore[override]
        rval: None = super().write()
        self.flush_to_index()
        return rval

    # END overridden methods


# } END classes



"""
BatchManager — the master-side driver for async batch jobs.

A ``SignalHandlingProcess`` started by the master's
``ProcessManager`` alongside ``Maintenance``, the Reactor engine, and
the rest of the master process tree.  Idles on ``get_event(wait=...)``
when no async batches are active; wakes on minion returns for adopted
batch JIDs and on ``salt/batch/*`` lifecycle events.

Layering:

- :mod:`salt.utils.batch_state` owns the pure state machine.
- :mod:`salt.utils.batch_output` owns the event tag vocabulary and
  payload shapes.
- This module composes them with Salt's event bus and ``LocalClient``.

The class is deliberately driven by small public-ish methods
(``_handle_event``, ``_progress_one``, ``_tick``, ``_handle_new``,
``_handle_stop``, ``_handle_recover``) so unit tests can step through
a scenario without a real master or multiprocessing fork.
"""

import logging
import re
import time

import salt.client
import salt.utils.batch_output
import salt.utils.batch_state
import salt.utils.event
import salt.utils.process

log = logging.getLogger(__name__)


_BATCH_LIFECYCLE_TAG = re.compile(r"^salt/batch/([^/]+)/(new|stop|recover)$")
_JOB_RETURN_TAG = re.compile(r"^salt/job/([^/]+)/ret/([^/]+)$")

DEFAULT_LOOP_INTERVAL = 5


class BatchManager(salt.utils.process.SignalHandlingProcess):
    """
    Master-side driver for async batch jobs.

    Listens for:

    - ``salt/job/<jid>/ret/<minion>`` — minion returns for active batches.
    - ``salt/batch/<jid>/new`` — new async batch registration from the CLI.
    - ``salt/batch/<jid>/stop`` — manual batch stop request from the
      ``batch.stop`` runner.
    - ``salt/batch/<jid>/recover`` — recovery request from the
      Maintenance safety net.

    On each relevant event, reads ``BatchState`` from ``.batch.p``,
    calls :func:`salt.utils.batch_state.progress_batch` to advance the
    state machine, publishes the next sub-batch if needed, writes the
    updated state back, and fires lifecycle events via the
    :class:`salt.utils.batch_output.EventOutput` adapter.

    Maintains an in-memory ``set`` of active batch JIDs so non-batch
    return events can be filtered with a single dict lookup — zero
    disk I/O on the hot path.
    """

    def __init__(self, opts, **kwargs):
        """
        :param dict opts: The salt master options dictionary.
        """
        super().__init__(**kwargs)
        self.opts = opts
        # Populated in _post_fork_init():
        self.event = None
        self.local = None
        self.output = None
        self.active_batches = set()

    # ------------------------------------------------------------------
    # Lifecycle — forked-process initialization and main loop
    # ------------------------------------------------------------------

    def _post_fork_init(self):
        """
        Initialize resources that must be created after the fork.

        Opens the master event bus with ``listen=True``, instantiates a
        ``LocalClient`` for sub-batch publishes, loads the active-batch
        index from disk, and builds the ``EventOutput`` adapter.
        """
        self.event = salt.utils.event.get_master_event(
            self.opts, self.opts["sock_dir"], listen=True
        )
        self.local = salt.client.get_local_client(
            self.opts.get("conf_file"), listen=False
        )
        self.active_batches = salt.utils.batch_state.read_active_index(self.opts)
        self.output = salt.utils.batch_output.EventOutput(self.opts, self.event)
        log.info(
            "BatchManager initialized with %d active batch(es) in the index",
            len(self.active_batches),
        )

    def run(self):
        """
        Main event loop.

        Blocks on ``get_event(wait=N)`` with N set by
        ``batch_manager_loop_interval`` (default 5s).  On every
        iteration — whether or not an event arrived — runs a housekeeping
        ``_tick()`` pass so timeout detection and batch_wait expiry
        progress even when the bus is quiet.
        """
        self._post_fork_init()
        loop_interval = self.opts.get(
            "batch_manager_loop_interval", DEFAULT_LOOP_INTERVAL
        )
        while True:
            try:
                event = self.event.get_event(
                    wait=loop_interval, full=True, no_block=False
                )
            except Exception:  # pylint: disable=broad-except
                log.exception("BatchManager failed to read from the event bus")
                event = None

            if event is not None:
                try:
                    self._handle_event(event)
                except Exception:  # pylint: disable=broad-except
                    log.exception(
                        "BatchManager failed to handle event %r", event.get("tag")
                    )
            try:
                self._tick()
            except Exception:  # pylint: disable=broad-except
                log.exception("BatchManager housekeeping tick failed")

    # ------------------------------------------------------------------
    # Event dispatch
    # ------------------------------------------------------------------

    def _handle_event(self, event):
        """
        Dispatch an event from the master bus to the right handler.

        :param dict event: Event dict with ``tag`` and ``data`` keys,
            as produced by ``SaltEvent.get_event(full=True)``.
        """
        tag = event.get("tag", "") if isinstance(event, dict) else ""
        data = event.get("data", {}) if isinstance(event, dict) else {}

        lifecycle = _BATCH_LIFECYCLE_TAG.match(tag)
        if lifecycle:
            jid, action = lifecycle.group(1), lifecycle.group(2)
            if action == "new":
                self._handle_new(jid)
            elif action == "stop":
                self._handle_stop(jid, data)
            elif action == "recover":
                self._handle_recover(jid)
            return

        ret = _JOB_RETURN_TAG.match(tag)
        if ret:
            jid, minion_id = ret.group(1), ret.group(2)
            if jid in self.active_batches:
                self._handle_batch_return(jid, minion_id, data)

    # ------------------------------------------------------------------
    # Handlers
    # ------------------------------------------------------------------

    def _handle_new(self, jid):
        """
        Adopt a new async batch.

        The CLI writes ``.batch.p`` before firing ``salt/batch/<jid>/new``,
        so this handler reads the on-disk state to confirm and registers
        the JID in the in-memory active set.  Sync batches
        (``driver="cli"``) are ignored.
        """
        if jid in self.active_batches:
            return
        state = salt.utils.batch_state.read_batch_state(jid, self.opts)
        if state is None:
            log.warning("salt/batch/%s/new received but .batch.p is not readable", jid)
            return
        if state.get("driver") != "master":
            log.debug(
                "Ignoring salt/batch/%s/new — driver=%s is not master-driven",
                jid,
                state.get("driver"),
            )
            return
        self._adopt(jid)
        log.info("Adopted async batch %s on behalf of %s", jid, state.get("user"))

    def _handle_stop(self, jid, data):
        """
        Halt an active batch gracefully.

        Sets ``halted=True`` / ``halted_reason="stop"`` (or the reason
        carried in *data*) and fires ``salt/batch/<jid>/halted``.  The
        runner is responsible for the ``kill=True`` variant — that path
        publishes ``saltutil.kill_job`` against active minions *before*
        emitting ``salt/batch/<jid>/stop``.
        """
        state = salt.utils.batch_state.read_batch_state(jid, self.opts)
        if state is None:
            self._retire(jid)
            return
        if state.get("halted"):
            self._retire(jid)
            return
        state["halted"] = True
        state["halted_reason"] = data.get("reason") or "stop"
        state["last_progress"] = time.time()
        salt.utils.batch_state.write_batch_state(jid, state, self.opts)
        if self.output is not None:
            self.output.on_batch_halted(state)
        self._retire(jid)

    def _handle_recover(self, jid):
        """
        Re-adopt a batch that the Maintenance safety net flagged as stale.

        Reads the state from disk (since we may have crashed since it
        was last in-memory), adds it back to the active set, and forces
        a progress tick so any free slots dispatch immediately.
        """
        state = salt.utils.batch_state.read_batch_state(jid, self.opts)
        if state is None:
            self._retire(jid)
            return
        if state.get("halted"):
            self._retire(jid)
            return
        if state.get("driver") != "master":
            return
        self._adopt(jid)
        self._progress_one(jid, {})

    def _handle_batch_return(self, jid, minion_id, data):
        """
        Process a minion return event for an active batch.

        Translates the ``salt/job/<jid>/ret/<minion>`` event payload to
        the ``{minion: return_data}`` shape the state machine expects,
        then delegates to :meth:`_progress_one`.
        """
        self._progress_one(jid, {minion_id: _event_to_return(data)})

    # ------------------------------------------------------------------
    # Progress core — one step through the state machine for a given JID
    # ------------------------------------------------------------------

    def _progress_one(self, jid, new_returns, now=None):
        """
        Read ``.batch.p``, advance the state machine, persist the
        result, publish any new sub-batch, and fire progress events.
        """
        state = salt.utils.batch_state.read_batch_state(jid, self.opts)
        if state is None:
            log.warning(
                "Batch %s has no readable .batch.p; retiring from active set",
                jid,
            )
            self._retire(jid)
            return
        if state.get("halted"):
            self._retire(jid)
            return

        action = salt.utils.batch_state.progress_batch(state, new_returns, now=now)

        if action.publish:
            self._publish_sub_batch(state, action.publish)

        salt.utils.batch_state.write_batch_state(jid, state, self.opts)

        if action.publish or action.finished_minions or action.timed_out_minions:
            if self.output is not None:
                self.output.on_batch_progress(state)

        if salt.utils.batch_state.is_batch_done(state):
            if self.output is not None:
                self.output.on_batch_done(state)
            self._retire(jid)

    def _tick(self, now=None):
        """
        Housekeeping pass over every active batch.

        Three things happen, in order:

        1. Reconcile the in-memory active set with the on-disk index.
           Batches registered by other processes (the CLI at
           batch-creation time, the ``batch.stop`` runner, a previously
           crashed manager) can be adopted without requiring an
           event.  This closes the race where a ``salt/batch/<jid>/new``
           event is lost before we're listening.
        2. Advance each active batch by one tick — drives timeout
           detection and ``batch_wait`` expiry when no return events
           are arriving.
        3. ``_progress_one`` retires JIDs that are halted or missing
           from disk.
        """
        on_disk = salt.utils.batch_state.read_active_index(self.opts)
        for jid in on_disk - self.active_batches:
            log.info(
                "BatchManager adopting batch %s discovered via active index",
                jid,
            )
            self.active_batches.add(jid)
        for jid in list(self.active_batches):
            self._progress_one(jid, {}, now=now)

    # ------------------------------------------------------------------
    # Side effects
    # ------------------------------------------------------------------

    def _publish_sub_batch(self, state, minion_ids):
        """
        Publish the next sub-batch via ``LocalClient.run_job``.

        ``jid`` is forced to the batch JID so every sub-batch publish
        accumulates under the same job directory (the single-JID
        foundation from PR #68941).  ``user`` is carried from
        ``BatchState`` so publisher-ACL enforcement and audit still
        attribute the publish to the original operator.
        """
        log.info(
            "Dispatching sub-batch of %d minion(s) for batch %s on behalf of %s",
            len(minion_ids),
            state.get("jid"),
            state.get("user"),
        )
        if self.local is None:
            log.error(
                "BatchManager has no LocalClient; cannot publish sub-batch for %s",
                state.get("jid"),
            )
            return
        try:
            self.local.run_job(
                tgt=list(minion_ids),
                fun=state["fun"],
                arg=list(state.get("arg") or []),
                tgt_type="list",
                ret=state.get("ret") or "",
                timeout=state.get("timeout", 60),
                jid=state["jid"],
                kwarg=dict(state.get("kwargs") or {}),
                listen=False,
                user=state.get("user", "root"),
            )
        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to publish sub-batch for %s", state.get("jid"))

    # ------------------------------------------------------------------
    # Active-set management
    # ------------------------------------------------------------------

    def _adopt(self, jid):
        """Add *jid* to the active set and persist the index."""
        self.active_batches.add(jid)
        salt.utils.batch_state.add_to_active_index(jid, self.opts)

    def _retire(self, jid):
        """Remove *jid* from the active set and persist the index."""
        self.active_batches.discard(jid)
        salt.utils.batch_state.remove_from_active_index(jid, self.opts)


# ---------------------------------------------------------------------------
# Event → return-payload translation
# ---------------------------------------------------------------------------


def _event_to_return(event_data):
    """
    Map a ``salt/job/<jid>/ret/<minion>`` event payload to the dict
    shape that :func:`salt.utils.batch_state.progress_batch` expects
    (identical to sync batch's ``parts[minion]``).

    The return-event schema uses ``return`` for the payload; the state
    machine follows sync batch and calls the key ``ret``.  ``retcode``
    is passed through unchanged (the state machine handles dict vs int
    collapse internally).
    """
    if not isinstance(event_data, dict):
        return {"ret": event_data, "retcode": 0}
    return {
        "ret": event_data.get("return"),
        "retcode": event_data.get("retcode", 0),
    }



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import hmac
import time
import base64
from typing import Dict, Optional
from hashlib import sha1
from datetime import datetime

import libcloud.utils.py3
from libcloud.utils.py3 import b, httplib, tostring, urlquote, urlencode
from libcloud.utils.xml import findtext, fixxpath
from libcloud.common.aws import (
    AWSDriver,
    AWSBaseResponse,
    AWSTokenConnection,
    SignedAWSConnection,
    UnsignedPayloadSentinel,
)
from libcloud.common.base import RawResponse, ConnectionUserAndKey
from libcloud.utils.files import read_in_chunks
from libcloud.common.types import LibcloudError, InvalidCredsError
from libcloud.storage.base import Object, Container, StorageDriver
from libcloud.storage.types import (
    ContainerError,
    ObjectDoesNotExistError,
    ObjectHashMismatchError,
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)

try:
    if libcloud.utils.py3.DEFAULT_LXML:
        from lxml.etree import Element, SubElement
    else:
        from xml.etree.ElementTree import Element, SubElement
except ImportError:
    from xml.etree.ElementTree import Element, SubElement


# How long before the token expires
EXPIRATION_SECONDS = 15 * 60

S3_US_STANDARD_HOST = "s3.amazonaws.com"
S3_US_EAST2_HOST = "s3-us-east-2.amazonaws.com"
S3_US_WEST_HOST = "s3-us-west-1.amazonaws.com"
S3_US_WEST_OREGON_HOST = "s3-us-west-2.amazonaws.com"
S3_US_GOV_EAST_HOST = "s3-us-gov-east-1.amazonaws.com"
S3_US_GOV_WEST_HOST = "s3-us-gov-west-1.amazonaws.com"
S3_CN_NORTH_HOST = "s3.cn-north-1.amazonaws.com.cn"
S3_CN_NORTHWEST_HOST = "s3.cn-northwest-1.amazonaws.com.cn"
S3_EU_WEST_HOST = "s3-eu-west-1.amazonaws.com"
S3_EU_WEST2_HOST = "s3-eu-west-2.amazonaws.com"
S3_EU_WEST3_HOST = "s3-eu-west-3.amazonaws.com"
S3_EU_CENTRAL_HOST = "s3-eu-central-1.amazonaws.com"
S3_EU_NORTH1_HOST = "s3-eu-north-1.amazonaws.com"
S3_EU_SOUTH1_HOST = "s3-eu-south-1.amazonaws.com"
S3_AP_SOUTH_HOST = "s3-ap-south-1.amazonaws.com"
S3_AP_SOUTHEAST_HOST = "s3-ap-southeast-1.amazonaws.com"
S3_AP_SOUTHEAST2_HOST = "s3-ap-southeast-2.amazonaws.com"
S3_AP_NORTHEAST1_HOST = "s3-ap-northeast-1.amazonaws.com"
S3_AP_NORTHEAST2_HOST = "s3-ap-northeast-2.amazonaws.com"
S3_AP_NORTHEAST_HOST = S3_AP_NORTHEAST1_HOST
S3_SA_EAST_HOST = "s3-sa-east-1.amazonaws.com"
S3_SA_SOUTHEAST2_HOST = "s3-sa-east-2.amazonaws.com"
S3_CA_CENTRAL_HOST = "s3-ca-central-1.amazonaws.com"
S3_AF_SOUTH1_HOST = "s3.af-south-1.amazonaws.com"

# Maps AWS region name to connection hostname
REGION_TO_HOST_MAP = {
    "us-east-1": S3_US_STANDARD_HOST,
    "us-east-2": S3_US_EAST2_HOST,
    "us-west-1": S3_US_WEST_HOST,
    "us-west-2": S3_US_WEST_OREGON_HOST,
    "us-gov-east-1": S3_US_GOV_EAST_HOST,
    "us-gov-west-1": S3_US_GOV_WEST_HOST,
    "cn-north-1": S3_CN_NORTH_HOST,
    "cn-northwest-1": S3_CN_NORTHWEST_HOST,
    "eu-west-1": S3_EU_WEST_HOST,
    "eu-west-2": S3_EU_WEST2_HOST,
    "eu-west-3": S3_EU_WEST3_HOST,
    "eu-north-1": "s3.eu-north-1.amazonaws.com",
    "eu-south-1": "s3.eu-south-1.amazonaws.com",
    "eu-central-1": S3_EU_CENTRAL_HOST,
    "ap-south-1": S3_AP_SOUTH_HOST,
    "ap-southeast-1": S3_AP_SOUTHEAST_HOST,
    "ap-southeast-2": S3_AP_SOUTHEAST2_HOST,
    "ap-northeast-1": S3_AP_NORTHEAST1_HOST,
    "ap-northeast-2": S3_AP_NORTHEAST2_HOST,
    "ap-northeast-3": "s3.ap-northeast-3.amazonaws.com",
    "sa-east-1": S3_SA_EAST_HOST,
    "sa-east-2": S3_SA_SOUTHEAST2_HOST,
    "ca-central-1": S3_CA_CENTRAL_HOST,
    "me-south-1": "s3.me-south-1.amazonaws.com",
    "af-south-1": S3_AF_SOUTH1_HOST,
    "me-central-1": "s3.me-central-1.amazonaws.com",
}

API_VERSION = "2006-03-01"
NAMESPACE = "http://s3.amazonaws.com/doc/%s/" % (API_VERSION)

# AWS multi-part chunks must be minimum 5MB
CHUNK_SIZE = 5 * 1024 * 1024

# Desired number of items in each response inside a paginated request in
# ex_iterate_multipart_uploads.
RESPONSES_PER_REQUEST = 100

S3_CDN_URL_DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"
S3_CDN_URL_DATE_FORMAT = "%Y%m%d"
S3_CDN_URL_EXPIRY_HOURS = float(os.getenv("LIBCLOUD_S3_CDN_URL_EXPIRY_HOURS", "24"))


class S3Response(AWSBaseResponse):
    namespace = None
    valid_response_codes = [
        httplib.NOT_FOUND,
        httplib.CONFLICT,
        httplib.BAD_REQUEST,
        httplib.PARTIAL_CONTENT,
    ]

    def success(self):
        i = int(self.status)
        return 200 <= i <= 299 or i in self.valid_response_codes

    def parse_error(self):
        if self.status in [httplib.UNAUTHORIZED, httplib.FORBIDDEN]:
            raise InvalidCredsError(self.body)
        elif self.status == httplib.MOVED_PERMANENTLY:
            bucket_region = self.headers.get("x-amz-bucket-region", None)
            used_region = self.connection.driver.region
            raise LibcloudError(
                "This bucket is located in a different "
                "region. Please use the correct driver. "
                'Bucket region "%s", used region "%s".' % (bucket_region, used_region),
                driver=S3StorageDriver,
            )
        raise LibcloudError(
            "Unknown error. Status code: %d" % (self.status), driver=S3StorageDriver
        )


class S3RawResponse(S3Response, RawResponse):
    pass


class BaseS3Connection(ConnectionUserAndKey):
    """
    Represents a single connection to the S3 Endpoint
    """

    host = "s3.amazonaws.com"
    responseCls = S3Response
    rawResponseCls = S3RawResponse

    @staticmethod
    def get_auth_signature(method, headers, params, expires, secret_key, path, vendor_prefix):
        """
        Signature = URL-Encode( Base64( HMAC-SHA1( YourSecretAccessKeyID,
                                    UTF-8-Encoding-Of( StringToSign ) ) ) );

        StringToSign = HTTP-VERB + "\n" +
            Content-MD5 + "\n" +
            Content-Type + "\n" +
            Expires + "\n" +
            CanonicalizedVendorHeaders +
            CanonicalizedResource;
        """
        special_headers = {"content-md5": "", "content-type": "", "date": ""}
        vendor_headers = {}

        for key, value in list(headers.items()):
            key_lower = key.lower()
            if key_lower in special_headers:
                special_headers[key_lower] = value.strip()
            elif key_lower.startswith(vendor_prefix):
                vendor_headers[key_lower] = value.strip()

        if expires:
            special_headers["date"] = str(expires)

        buf = [method]
        for _, value in sorted(special_headers.items()):
            buf.append(value)
        string_to_sign = "\n".join(buf)

        buf = []
        for key, value in sorted(vendor_headers.items()):
            buf.append("{}:{}".format(key, value))
        header_string = "\n".join(buf)

        values_to_sign = []
        for value in [string_to_sign, header_string, path]:
            if value:
                values_to_sign.append(value)

        string_to_sign = "\n".join(values_to_sign)
        b64_hmac = base64.b64encode(
            hmac.new(b(secret_key), b(string_to_sign), digestmod=sha1).digest()
        )
        return b64_hmac.decode("utf-8")

    def add_default_params(self, params):
        expires = str(int(time.time()) + EXPIRATION_SECONDS)
        params["AWSAccessKeyId"] = self.user_id
        params["Expires"] = expires
        return params

    def pre_connect_hook(self, params, headers):
        # pylint: disable=no-member
        params["Signature"] = self.get_auth_signature(
            method=self.method,
            headers=headers,
            params=params,
            expires=params["Expires"],
            secret_key=self.key,
            path=self.action,
            vendor_prefix=self.driver.http_vendor_prefix,
        )
        return params, headers


class S3Connection(AWSTokenConnection, BaseS3Connection):
    """
    Represents a single connection to the S3 endpoint, with AWS-specific
    features.
    """

    pass


class S3SignatureV4Connection(SignedAWSConnection, BaseS3Connection):
    service_name = "s3"
    version = API_VERSION

    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        token=None,
        retry_delay=None,
        backoff=None,
    ):
        super().__init__(
            user_id,
            key,
            secure,
            host,
            port,
            url,
            timeout,
            proxy_url,
            token,
            retry_delay,
            backoff,
            4,
        )  # force version 4


class S3MultipartUpload:
    """
    Class representing an amazon s3 multipart upload
    """

    def __init__(self, key, id, created_at, initiator, owner):
        """
        Class representing an amazon s3 multipart upload

        :param key: The object/key that was being uploaded
        :type key: ``str``

        :param id: The upload id assigned by amazon
        :type id: ``str``

        :param created_at: The date/time at which the upload was started
        :type created_at: ``str``

        :param initiator: The AWS owner/IAM user who initiated this
        :type initiator: ``str``

        :param owner: The AWS owner/IAM who will own this object
        :type owner: ``str``
        """
        self.key = key
        self.id = id
        self.created_at = created_at
        self.initiator = initiator
        self.owner = owner

    def __repr__(self):
        return "<S3MultipartUpload: key=%s>" % (self.key)


class BaseS3StorageDriver(StorageDriver):
    name = "Amazon S3 (standard)"
    website = "http://aws.amazon.com/s3/"
    connectionCls = BaseS3Connection
    hash_type = "md5"
    supports_chunked_encoding = False
    supports_s3_multipart_upload = True
    ex_location_name = ""
    namespace = NAMESPACE
    http_vendor_prefix = "x-amz"

    def iterate_containers(self):
        response = self.connection.request("/")
        if response.status == httplib.OK:
            containers = self._to_containers(obj=response.object, xpath="Buckets/Bucket")
            return containers

        raise LibcloudError("Unexpected status code: %s" % (response.status), driver=self)

    def iterate_container_objects(self, container, prefix=None, ex_prefix=None):
        """
        Return a generator of objects for the given container.

        :param container: Container instance
        :type container: :class:`Container`

        :param prefix: Only return objects starting with prefix
        :type prefix: ``str``

        :param ex_prefix: Only return objects starting with ex_prefix
        :type ex_prefix: ``str``

        :return: A generator of Object instances.
        :rtype: ``generator`` of :class:`Object`
        """
        prefix = self._normalize_prefix_argument(prefix, ex_prefix)

        params = {}

        if prefix:
            params["prefix"] = prefix

        last_key = None
        exhausted = False
        container_path = self._get_container_path(container)

        while not exhausted:
            if last_key:
                params["marker"] = last_key

            response = self.connection.request(container_path, params=params)

            if response.status != httplib.OK:
                raise LibcloudError("Unexpected status code: %s" % (response.status), driver=self)

            objects = self._to_objs(obj=response.object, xpath="Contents", container=container)
            is_truncated = response.object.findtext(
                fixxpath(xpath="IsTruncated", namespace=self.namespace)
            ).lower()
            exhausted = is_truncated == "false"

            last_key = None
            for obj in objects:
                last_key = obj.name
                yield obj

    def get_container(self, container_name):
        try:
            response = self.connection.request("/%s" % container_name, method="HEAD")
            if response.status == httplib.NOT_FOUND:
                raise ContainerDoesNotExistError(
                    value=None, driver=self, container_name=container_name
                )
        except InvalidCredsError:
            # This just means the user doesn't have IAM permissions to do a
            # HEAD request but other requests might work.
            pass
        return Container(name=container_name, extra=None, driver=self)

    def get_object(self, container_name, object_name):
        container = self.get_container(container_name=container_name)
        object_path = self._get_object_path(container, object_name)
        response = self.connection.request(object_path, method="HEAD")

        if response.status == httplib.OK:
            obj = self._headers_to_object(
                object_name=object_name, container=container, headers=response.headers
            )
            return obj

        raise ObjectDoesNotExistError(value=None, driver=self, object_name=object_name)

    def _get_container_path(self, container):
        """
        Return a container path

        :param container: Container instance
        :type  container: :class:`Container`

        :return: A path for this container.
        :rtype: ``str``
        """
        return "/%s" % (container.name)

    def _get_object_path(self, container, object_name):
        """
        Return an object's CDN path.

        :param container: Container instance
        :type  container: :class:`Container`

        :param object_name: Object name
        :type  object_name: :class:`str`

        :return: A  path for this object.
        :rtype: ``str``
        """
        container_url = self._get_container_path(container)
        object_name_cleaned = self._clean_object_name(object_name)
        object_path = "{}/{}".format(container_url, object_name_cleaned)
        return object_path

    def create_container(self, container_name):
        if self.ex_location_name:
            root = Element("CreateBucketConfiguration")
            child = SubElement(root, "LocationConstraint")
            child.text = self.ex_location_name

            data = tostring(root)
        else:
            data = ""

        response = self.connection.request("/%s" % (container_name), data=data, method="PUT")

        if response.status == httplib.OK:
            container = Container(name=container_name, extra=None, driver=self)
            return container
        elif response.status == httplib.CONFLICT:
            if "BucketAlreadyOwnedByYou" in response.body:
                raise ContainerAlreadyExistsError(
                    value="Container with this name already exists. The name "
                    "be unique among all the containers in the system.",
                    container_name=container_name,
                    driver=self,
                )

            raise InvalidContainerNameError(
                value="Container with this name already exists. The name must "
                "be unique among all the containers in the system.",
                container_name=container_name,
                driver=self,
            )
        elif response.status == httplib.BAD_REQUEST:
            raise ContainerError(
                value="Bad request when creating container: %s" % response.body,
                container_name=container_name,
                driver=self,
            )

        raise LibcloudError("Unexpected status code: %s" % (response.status), driver=self)

    def delete_container(self, container):
        # Note: All the objects in the container must be deleted first
        response = self.connection.request("/%s" % (container.name), method="DELETE")
        if response.status == httplib.NO_CONTENT:
            return True
        elif response.status == httplib.CONFLICT:
            raise ContainerIsNotEmptyError(
                value="Container must be empty before it can be deleted.",
                container_name=container.name,
                driver=self,
            )
        elif response.status == httplib.NOT_FOUND:
            raise ContainerDoesNotExistError(value=None, driver=self, container_name=container.name)

        return False

    def download_object(
        self, obj, destination_path, overwrite_existing=False, delete_on_failure=True
    ):
        obj_path = self._get_object_path(obj.container, obj.name)

        response = self.connection.request(obj_path, method="GET", raw=True)

        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
            },
            success_status_code=httplib.OK,
        )

    def download_object_as_stream(self, obj, chunk_size=None):
        obj_path = self._get_object_path(obj.container, obj.name)
        response = self.connection.request(obj_path, method="GET", stream=True, raw=True)

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={
                "iterator": response.iter_content(CHUNK_SIZE),
                "chunk_size": chunk_size,
            },
            success_status_code=httplib.OK,
        )

    def download_object_range(
        self,
        obj,
        destination_path,
        start_bytes,
        end_bytes=None,
        overwrite_existing=False,
        delete_on_failure=True,
    ):
        self._validate_start_and_end_bytes(start_bytes=start_bytes, end_bytes=end_bytes)

        obj_path = self._get_object_path(obj.container, obj.name)

        headers = {"Range": self._get_standard_range_str(start_bytes, end_bytes)}
        response = self.connection.request(obj_path, method="GET", headers=headers, raw=True)

        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
                "partial_download": True,
            },
            success_status_code=httplib.PARTIAL_CONTENT,
        )

    def download_object_range_as_stream(self, obj, start_bytes, end_bytes=None, chunk_size=None):
        self._validate_start_and_end_bytes(start_bytes=start_bytes, end_bytes=end_bytes)

        obj_path = self._get_object_path(obj.container, obj.name)

        headers = {"Range": self._get_standard_range_str(start_bytes, end_bytes)}
        response = self.connection.request(
            obj_path, method="GET", headers=headers, stream=True, raw=True
        )

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={
                "iterator": response.iter_content(CHUNK_SIZE),
                "chunk_size": chunk_size,
            },
            success_status_code=httplib.PARTIAL_CONTENT,
        )

    def upload_object(
        self,
        file_path,
        container,
        object_name,
        extra=None,
        verify_hash=True,
        headers=None,
        ex_storage_class=None,
    ):
        """
        @inherits: :class:`StorageDriver.upload_object`

        :param ex_storage_class: Storage class
        :type ex_storage_class: ``str``
        """
        return self._put_object(
            container=container,
            object_name=object_name,
            extra=extra,
            file_path=file_path,
            verify_hash=verify_hash,
            headers=headers,
            storage_class=ex_storage_class,
        )

    def _initiate_multipart(self, container, object_name, headers=None):
        """
        Initiates a multipart upload to S3

        :param container: The destination container
        :type container: :class:`Container`

        :param object_name: The name of the object which we are uploading
        :type object_name: ``str``

        :keyword headers: Additional headers to send with the request
        :type headers: ``dict``

        :return: The id of the newly created multipart upload
        :rtype: ``str``
        """
        headers = headers or {}

        request_path = self._get_object_path(container, object_name)
        params = {"uploads": ""}

        response = self.connection.request(
            request_path, method="POST", headers=headers, params=params
        )

        if response.status != httplib.OK:
            raise LibcloudError("Error initiating multipart upload", driver=self)

        return findtext(element=response.object, xpath="UploadId", namespace=self.namespace)

    def _upload_multipart_chunks(
        self, container, object_name, upload_id, stream, calculate_hash=True
    ):
        """
        Uploads data from an iterator in fixed sized chunks to S3

        :param container: The destination container
        :type container: :class:`Container`

        :param object_name: The name of the object which we are uploading
        :type object_name: ``str``

        :param upload_id: The upload id allocated for this multipart upload
        :type upload_id: ``str``

        :param stream: The generator for fetching the upload data
        :type stream: ``generator``

        :keyword calculate_hash: Indicates if we must calculate the data hash
        :type calculate_hash: ``bool``

        :return: A tuple of (chunk info, checksum, bytes transferred)
        :rtype: ``tuple``
        """
        data_hash = None
        if calculate_hash:
            data_hash = self._get_hash_function()

        bytes_transferred = 0
        count = 1
        chunks = []
        params = {"uploadId": upload_id}

        request_path = self._get_object_path(container, object_name)

        # Read the input data in chunk sizes suitable for AWS
        for data in read_in_chunks(stream, chunk_size=CHUNK_SIZE, fill_size=True, yield_empty=True):
            bytes_transferred += len(data)

            if calculate_hash:
                data_hash.update(data)

            chunk_hash = self._get_hash_function()
            chunk_hash.update(data)
            chunk_hash = base64.b64encode(chunk_hash.digest()).decode("utf-8")

            # The Content-MD5 header provides an extra level of data check and
            # is recommended by amazon
            headers = {
                "Content-Length": len(data),
                "Content-MD5": chunk_hash,
            }

            params["partNumber"] = count

            resp = self.connection.request(
                request_path, method="PUT", data=data, headers=headers, params=params
            )

            if resp.status != httplib.OK:
                raise LibcloudError("Error uploading chunk", driver=self)

            server_hash = resp.headers["etag"].replace('"', "")

            # Keep this data for a later commit
            chunks.append((count, server_hash))
            count += 1

        if calculate_hash:
            data_hash = data_hash.hexdigest()

        return (chunks, data_hash, bytes_transferred)

    def _commit_multipart(self, container, object_name, upload_id, chunks):
        """
        Makes a final commit of the data.

        :param container: The destination container
        :type container: :class:`Container`

        :param object_name: The name of the object which we are uploading
        :type object_name: ``str``

        :param upload_id: The upload id allocated for this multipart upload
        :type upload_id: ``str``

        :param chunks: A list of (chunk_number, chunk_hash) tuples.
        :type chunks: ``list``

        :return: The server side hash of the uploaded data
        :rtype: ``str``
        """
        root = Element("CompleteMultipartUpload")

        for count, etag in chunks:
            part = SubElement(root, "Part")
            part_no = SubElement(part, "PartNumber")
            part_no.text = str(count)

            etag_id = SubElement(part, "ETag")
            etag_id.text = str(etag)

        data = tostring(root)

        headers = {"Content-Length": len(data)}
        params = {"uploadId": upload_id}
        request_path = self._get_object_path(container, object_name)
        response = self.connection.request(
            request_path, headers=headers, params=params, data=data, method="POST"
        )

        if response.status != httplib.OK:
            element = response.object
            # pylint: disable=maybe-no-member
            code, message = response._parse_error_details(element=element)
            msg = "Error in multipart commit: {} ({})".format(message, code)
            raise LibcloudError(msg, driver=self)

        # Get the server's etag to be passed back to the caller
        body = response.parse_body()
        server_hash = body.find(fixxpath(xpath="ETag", namespace=self.namespace)).text
        return server_hash

    def _abort_multipart(self, container, object_name, upload_id):
        """
        Aborts an already initiated multipart upload

        :param container: The destination container
        :type container: :class:`Container`

        :param object_name: The name of the object which we are uploading
        :type object_name: ``str``

        :param upload_id: The upload id allocated for this multipart upload
        :type upload_id: ``str``
        """

        params = {"uploadId": upload_id}
        request_path = self._get_object_path(container, object_name)

        resp = self.connection.request(request_path, method="DELETE", params=params)

        if resp.status != httplib.NO_CONTENT:
            raise LibcloudError(
                "Error in multipart abort. status_code=%d" % (resp.status), driver=self
            )

    def upload_object_via_stream(
        self,
        iterator,
        container,
        object_name,
        extra=None,
        headers=None,
        ex_storage_class=None,
    ):
        """
        @inherits: :class:`StorageDriver.upload_object_via_stream`

        :param ex_storage_class: Storage class
        :type ex_storage_class: ``str``
        """

        method = "PUT"
        params = None

        # This driver is used by other S3 API compatible drivers also.
        # Amazon provides a different (complex?) mechanism to do multipart
        # uploads
        if self.supports_s3_multipart_upload:
            return self._put_object_multipart(
                container=container,
                object_name=object_name,
                extra=extra,
                stream=iterator,
                verify_hash=False,
                headers=headers,
                storage_class=ex_storage_class,
            )
        return self._put_object(
            container=container,
            object_name=object_name,
            extra=extra,
            method=method,
            query_args=params,
            stream=iterator,
            verify_hash=False,
            headers=headers,
            storage_class=ex_storage_class,
        )

    def delete_object(self, obj):
        object_path = self._get_object_path(obj.container, obj.name)
        response = self.connection.request(object_path, method="DELETE")
        if response.status == httplib.NO_CONTENT:
            return True
        elif response.status == httplib.NOT_FOUND:
            raise ObjectDoesNotExistError(value=None, driver=self, object_name=obj.name)

        return False

    def ex_iterate_multipart_uploads(self, container, prefix=None, delimiter=None):
        """
        Extension method for listing all in-progress S3 multipart uploads.

        Each multipart upload which has not been committed or aborted is
        considered in-progress.

        :param container: The container holding the uploads
        :type container: :class:`Container`

        :keyword prefix: Print only uploads of objects with this prefix
        :type prefix: ``str``

        :keyword delimiter: The object/key names are grouped based on
            being split by this delimiter
        :type delimiter: ``str``

        :return: A generator of S3MultipartUpload instances.
        :rtype: ``generator`` of :class:`S3MultipartUpload`
        """

        if not self.supports_s3_multipart_upload:
            raise LibcloudError("Feature not supported", driver=self)

        # Get the data for a specific container
        request_path = self._get_container_path(container)
        params = {"max-uploads": RESPONSES_PER_REQUEST, "uploads": ""}

        if prefix:
            params["prefix"] = prefix

        if delimiter:
            params["delimiter"] = delimiter

        def finder(node, text):
            return node.findtext(fixxpath(xpath=text, namespace=self.namespace))

        while True:
            response = self.connection.request(request_path, params=params)

            if response.status != httplib.OK:
                raise LibcloudError(
                    "Error fetching multipart uploads. " "Got code: %s" % response.status,
                    driver=self,
                )

            body = response.parse_body()
            # pylint: disable=maybe-no-member
            for node in body.findall(fixxpath(xpath="Upload", namespace=self.namespace)):
                initiator = node.find(fixxpath(xpath="Initiator", namespace=self.namespace))
                owner = node.find(fixxpath(xpath="Owner", namespace=self.namespace))

                key = finder(node, "Key")
                upload_id = finder(node, "UploadId")
                created_at = finder(node, "Initiated")
                initiator = finder(initiator, "DisplayName")
                owner = finder(owner, "DisplayName")

                yield S3MultipartUpload(key, upload_id, created_at, initiator, owner)

            # Check if this is the last entry in the listing
            # pylint: disable=maybe-no-member
            is_truncated = body.findtext(fixxpath(xpath="IsTruncated", namespace=self.namespace))

            if is_truncated.lower() == "false":
                break

            # Provide params for the next request
            upload_marker = body.findtext(
                fixxpath(xpath="NextUploadIdMarker", namespace=self.namespace)
            )
            key_marker = body.findtext(fixxpath(xpath="NextKeyMarker", namespace=self.namespace))

            params["key-marker"] = key_marker
            params["upload-id-marker"] = upload_marker

    def ex_cleanup_all_multipart_uploads(self, container, prefix=None):
        """
        Extension method for removing all partially completed S3 multipart
        uploads.

        :param container: The container holding the uploads
        :type container: :class:`Container`

        :keyword prefix: Delete only uploads of objects with this prefix
        :type prefix: ``str``
        """

        # Iterate through the container and delete the upload ids
        for upload in self.ex_iterate_multipart_uploads(container, prefix, delimiter=None):
            self._abort_multipart(container, upload.key, upload.id)

    def _clean_object_name(self, name):
        name = urlquote(name, safe="/~")
        return name

    def _put_object(
        self,
        container,
        object_name,
        method="PUT",
        query_args=None,
        extra=None,
        file_path=None,
        stream=None,
        verify_hash=True,
        storage_class=None,
        headers=None,
    ):
        headers = headers or {}
        extra = extra or {}

        headers.update(self._to_storage_class_headers(storage_class))

        content_type = extra.get("content_type", None)
        meta_data = extra.get("meta_data", None)
        acl = extra.get("acl", None)

        if meta_data:
            for key, value in list(meta_data.items()):
                key = self.http_vendor_prefix + "-meta-%s" % (key)
                headers[key] = value

        if acl:
            headers[self.http_vendor_prefix + "-acl"] = acl

        request_path = self._get_object_path(container, object_name)

        if query_args:
            request_path = "?".join((request_path, query_args))

        result_dict = self._upload_object(
            object_name=object_name,
            content_type=content_type,
            request_path=request_path,
            request_method=method,
            headers=headers,
            file_path=file_path,
            stream=stream,
        )

        response = result_dict["response"]
        bytes_transferred = result_dict["bytes_transferred"]
        headers = response.headers
        response = response
        server_hash = headers.get("etag", "").replace('"', "")
        server_side_encryption = headers.get("x-amz-server-side-encryption", None)
        aws_kms_encryption = server_side_encryption == "aws:kms"
        hash_matches = result_dict["data_hash"] == server_hash

        # NOTE: If AWS KMS server side encryption is enabled, ETag won't
        # contain object MD5 digest so we skip the checksum check
        # See https://docs.aws.amazon.com/AmazonS3/latest/API
        # /RESTCommonResponseHeaders.html
        # and https://github.com/apache/libcloud/issues/1401
        # for details
        if verify_hash and not aws_kms_encryption and not hash_matches:
            raise ObjectHashMismatchError(
                value="MD5 hash {} checksum does not match {}".format(
                    server_hash, result_dict["data_hash"]
                ),
                object_name=object_name,
                driver=self,
            )
        elif response.status == httplib.OK:
            obj = Object(
                name=object_name,
                size=bytes_transferred,
                hash=server_hash,
                extra={"acl": acl},
                meta_data=meta_data,
                container=container,
                driver=self,
            )

            return obj
        else:
            raise LibcloudError(
                "Unexpected status code, status_code=%s" % (response.status),
                driver=self,
            )

    def _put_object_multipart(
        self,
        container,
        object_name,
        stream,
        extra=None,
        verify_hash=False,
        headers=None,
        storage_class=None,
    ):
        """
        Uploads an object using the S3 multipart algorithm.

        :param container: The destination container
        :type container: :class:`Container`

        :param object_name: The name of the object which we are uploading
        :type object_name: ``str``

        :param stream: The generator for fetching the upload data
        :type stream: ``generator``

        :keyword verify_hash: Indicates if we must calculate the data hash
        :type verify_hash: ``bool``

        :keyword extra: Additional options
        :type extra: ``dict``

        :keyword headers: Additional headers
        :type headers: ``dict``

        :keyword storage_class: The name of the S3 object's storage class
        :type extra: ``str``

        :return: The uploaded object
        :rtype: :class:`Object`
        """
        headers = headers or {}
        extra = extra or {}

        headers.update(self._to_storage_class_headers(storage_class))

        content_type = extra.get("content_type", None)
        meta_data = extra.get("meta_data", None)
        acl = extra.get("acl", None)

        headers["Content-Type"] = self._determine_content_type(content_type, object_name)

        if meta_data:
            for key, value in list(meta_data.items()):
                key = self.http_vendor_prefix + "-meta-%s" % (key)
                headers[key] = value

        if acl:
            headers[self.http_vendor_prefix + "-acl"] = acl

        upload_id = self._initiate_multipart(container, object_name, headers=headers)

        try:
            result = self._upload_multipart_chunks(
                container, object_name, upload_id, stream, calculate_hash=verify_hash
            )
            chunks, data_hash, bytes_transferred = result

            # Commit the chunk info and complete the upload
            etag = self._commit_multipart(container, object_name, upload_id, chunks)
        except Exception:
            # Amazon provides a mechanism for aborting an upload.
            self._abort_multipart(container, object_name, upload_id)
            raise

        return Object(
            name=object_name,
            size=bytes_transferred,
            hash=etag,
            extra={"acl": acl},
            meta_data=meta_data,
            container=container,
            driver=self,
        )

    def _to_storage_class_headers(self, storage_class):
        """
        Generates request headers given a storage class name.

        :keyword storage_class: The name of the S3 object's storage class
        :type extra: ``str``

        :return: Headers to include in a request
        :rtype: :dict:
        """
        headers = {}
        storage_class = storage_class or "standard"
        if storage_class not in [
            "standard",
            "reduced_redundancy",
            "standard_ia",
            "onezone_ia",
            "intelligent_tiering",
            "glacier",
            "deep_archive",
            "glacier_ir",
        ]:
            raise ValueError("Invalid storage class value: %s" % (storage_class))

        key = self.http_vendor_prefix + "-storage-class"
        headers[key] = storage_class.upper()
        return headers

    def _to_containers(self, obj, xpath):
        for element in obj.findall(fixxpath(xpath=xpath, namespace=self.namespace)):
            yield self._to_container(element)

    def _to_objs(self, obj, xpath, container):
        return [
            self._to_obj(element, container)
            for element in obj.findall(fixxpath(xpath=xpath, namespace=self.namespace))
        ]

    def _to_container(self, element):
        extra = {
            "creation_date": findtext(
                element=element, xpath="CreationDate", namespace=self.namespace
            )
        }

        container = Container(
            name=findtext(element=element, xpath="Name", namespace=self.namespace),
            extra=extra,
            driver=self,
        )

        return container

    def _get_content_length_from_headers(self, headers: Dict[str, str]) -> Optional[int]:
        """
        Prase object size from the provided response headers.
        """
        content_length = headers.get("content-length", None)
        return content_length

    def _headers_to_object(self, object_name, container, headers):
        hash = headers.get("etag", "").replace('"', "")

        extra = {}

        # Not all the S3 compatible implementations return this header, see
        # https://github.com/apache/libcloud/pull/1695 for details
        if "content-type" in headers:
            extra["content_type"] = headers["content-type"]

        # Google Storage S3 compatible API doesn't return this header under
        # some scenarios https://github.com/apache/libcloud/issues/1682
        if "etag" in headers:
            extra["etag"] = headers["etag"]

        meta_data = {}

        if "content-encoding" in headers:
            extra["content_encoding"] = headers["content-encoding"]

        if "last-modified" in headers:
            extra["last_modified"] = headers["last-modified"]

        for key, value in headers.items():
            if not key.lower().startswith(self.http_vendor_prefix + "-meta-"):
                continue

            key = key.replace(self.http_vendor_prefix + "-meta-", "")
            meta_data[key] = value

        content_length = self._get_content_length_from_headers(headers=headers)

        if content_length is None:
            raise KeyError(
                "Can not deduce object size from headers for " "object %s" % (object_name)
            )

        obj = Object(
            name=object_name,
            size=int(content_length),
            hash=hash or None,
            extra=extra,
            meta_data=meta_data,
            container=container,
            driver=self,
        )
        return obj

    def _to_obj(self, element, container):
        owner_id = findtext(element=element, xpath="Owner/ID", namespace=self.namespace)
        owner_display_name = findtext(
            element=element, xpath="Owner/DisplayName", namespace=self.namespace
        )
        meta_data = {"owner": {"id": owner_id, "display_name": owner_display_name}}
        last_modified = findtext(element=element, xpath="LastModified", namespace=self.namespace)
        extra = {"last_modified": last_modified}

        obj = Object(
            name=findtext(element=element, xpath="Key", namespace=self.namespace),
            size=int(findtext(element=element, xpath="Size", namespace=self.namespace)),
            hash=findtext(element=element, xpath="ETag", namespace=self.namespace).replace('"', ""),
            extra=extra,
            meta_data=meta_data,
            container=container,
            driver=self,
        )

        return obj


class S3StorageDriver(AWSDriver, BaseS3StorageDriver):
    name = "Amazon S3"
    connectionCls = S3SignatureV4Connection
    region_name = "us-east-1"

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        region=None,
        token=None,
        **kwargs,
    ):
        # Here for backward compatibility for old and deprecated driver class
        # per region approach
        if hasattr(self, "region_name") and not region:
            region = self.region_name  # pylint: disable=no-member

        self.region_name = region

        if region and region not in REGION_TO_HOST_MAP.keys():
            raise ValueError("Invalid or unsupported region: %s" % (region))

        self.name = "Amazon S3 (%s)" % (region)

        if host is None:
            host = REGION_TO_HOST_MAP[region]

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            token=token,
            **kwargs,
        )

    @classmethod
    def list_regions(self):
        return REGION_TO_HOST_MAP.keys()

    def get_object_cdn_url(self, obj, ex_expiry=S3_CDN_URL_EXPIRY_HOURS):
        """
        Return a "presigned URL" for read-only access to object

        AWS only - requires AWS signature V4 authentication.

        :param obj: Object instance.
        :type  obj: :class:`Object`

        :param ex_expiry: The number of hours after which the URL expires.
                          Defaults to 24 hours or the value of the environment
                          variable "LIBCLOUD_S3_STORAGE_CDN_URL_EXPIRY_HOURS",
                          if set.
        :type  ex_expiry: ``float``

        :return: Presigned URL for the object.
        :rtype: ``str``
        """

        # assemble data for the request we want to pre-sign
        # see: https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-query-string-auth.html # noqa
        object_path = self._get_object_path(obj.container, obj.name)
        now = datetime.utcnow()
        duration_seconds = int(ex_expiry * 3600)
        credparts = (
            self.key,
            now.strftime(S3_CDN_URL_DATE_FORMAT),
            self.region,
            "s3",
            "aws4_request",
        )
        params_to_sign = {
            "X-Amz-Algorithm": "AWS4-HMAC-SHA256",
            "X-Amz-Credential": "/".join(credparts),
            "X-Amz-Date": now.strftime(S3_CDN_URL_DATETIME_FORMAT),
            "X-Amz-Expires": duration_seconds,
            "X-Amz-SignedHeaders": "host",
        }
        headers_to_sign = {"host": self.connection.host}

        # generate signature for the pre-signed request
        signature = self.connection.signer._get_signature(
            params=params_to_sign,
            headers=headers_to_sign,
            dt=now,
            method="GET",
            path=object_path,
            data=UnsignedPayloadSentinel,
        )

        # Create final params for pre-signed URL
        params = params_to_sign.copy()
        params["X-Amz-Signature"] = signature

        return "{scheme}://{host}:{port}{path}?{params}".format(
            scheme="https" if self.secure else "http",
            host=self.connection.host,
            port=self.connection.port,
            path=object_path,
            params=urlencode(params),
        )


class S3USEast2Connection(S3SignatureV4Connection):
    host = S3_US_EAST2_HOST


class S3USEast2StorageDriver(S3StorageDriver):
    name = "Amazon S3 (us-east-2)"
    connectionCls = S3USEast2Connection
    ex_location_name = "us-east-2"
    region_name = "us-east-2"


class S3USWestConnection(S3SignatureV4Connection):
    host = S3_US_WEST_HOST


class S3USWestStorageDriver(S3StorageDriver):
    name = "Amazon S3 (us-west-1)"
    connectionCls = S3USWestConnection
    ex_location_name = "us-west-1"
    region_name = "us-west-1"


class S3USWestOregonConnection(S3SignatureV4Connection):
    host = S3_US_WEST_OREGON_HOST


class S3USWestOregonStorageDriver(S3StorageDriver):
    name = "Amazon S3 (us-west-2)"
    connectionCls = S3USWestOregonConnection
    ex_location_name = "us-west-2"
    region_name = "us-west-2"


class S3USGovEastConnection(S3SignatureV4Connection):
    host = S3_US_GOV_EAST_HOST


class S3USGovEastStorageDriver(S3StorageDriver):
    name = "Amazon S3 (us-gov-east-1)"
    connectionCls = S3USGovEastConnection
    ex_location_name = "us-gov-east-1"
    region_name = "us-gov-east-1"


class S3USGovWestConnection(S3SignatureV4Connection):
    host = S3_US_GOV_WEST_HOST


class S3USGovWestStorageDriver(S3StorageDriver):
    name = "Amazon S3 (us-gov-west-1)"
    connectionCls = S3USGovWestConnection
    ex_location_name = "us-gov-west-1"
    region_name = "us-gov-west-1"


class S3CNNorthWestConnection(S3SignatureV4Connection):
    host = S3_CN_NORTHWEST_HOST


class S3CNNorthWestStorageDriver(S3StorageDriver):
    name = "Amazon S3 (cn-northwest-1)"
    connectionCls = S3CNNorthWestConnection
    ex_location_name = "cn-northwest-1"
    region_name = "cn-northwest-1"


class S3CNNorthConnection(S3SignatureV4Connection):
    host = S3_CN_NORTH_HOST


class S3CNNorthStorageDriver(S3StorageDriver):
    name = "Amazon S3 (cn-north-1)"
    connectionCls = S3CNNorthConnection
    ex_location_name = "cn-north-1"
    region_name = "cn-north-1"


class S3EUWestConnection(S3SignatureV4Connection):
    host = S3_EU_WEST_HOST


class S3EUWestStorageDriver(S3StorageDriver):
    name = "Amazon S3 (eu-west-1)"
    connectionCls = S3EUWestConnection
    ex_location_name = "EU"
    region_name = "eu-west-1"


class S3EUWest2Connection(S3SignatureV4Connection):
    host = S3_EU_WEST2_HOST


class S3EUWest2StorageDriver(S3StorageDriver):
    name = "Amazon S3 (eu-west-2)"
    connectionCls = S3EUWest2Connection
    ex_location_name = "eu-west-2"
    region_name = "eu-west-2"


class S3EUCentralConnection(S3SignatureV4Connection):
    host = S3_EU_CENTRAL_HOST


class S3EUCentralStorageDriver(S3StorageDriver):
    name = "Amazon S3 (eu-central-1)"
    connectionCls = S3EUCentralConnection
    ex_location_name = "eu-central-1"
    region_name = "eu-central-1"


class S3APSEConnection(S3SignatureV4Connection):
    host = S3_AP_SOUTHEAST_HOST


class S3EUNorth1Connection(S3SignatureV4Connection):
    host = S3_EU_NORTH1_HOST


class S3EUNorth1StorageDriver(S3StorageDriver):
    name = "Amazon S3 (eu-north-1)"
    connectionCls = S3EUNorth1Connection
    ex_location_name = "eu-north-1"
    region_name = "eu-north-1"


class S3APSEStorageDriver(S3StorageDriver):
    name = "Amazon S3 (ap-southeast-1)"
    connectionCls = S3APSEConnection
    ex_location_name = "ap-southeast-1"
    region_name = "ap-southeast-1"


class S3APSE2Connection(S3SignatureV4Connection):
    host = S3_AP_SOUTHEAST2_HOST


class S3APSE2StorageDriver(S3StorageDriver):
    name = "Amazon S3 (ap-southeast-2)"
    connectionCls = S3APSE2Connection
    ex_location_name = "ap-southeast-2"
    region_name = "ap-southeast-2"


class S3APNE1Connection(S3SignatureV4Connection):
    host = S3_AP_NORTHEAST1_HOST


S3APNEConnection = S3APNE1Connection


class S3APNE1StorageDriver(S3StorageDriver):
    name = "Amazon S3 (ap-northeast-1)"
    connectionCls = S3APNEConnection
    ex_location_name = "ap-northeast-1"
    region_name = "ap-northeast-1"


S3APNEStorageDriver = S3APNE1StorageDriver


class S3APNE2Connection(S3SignatureV4Connection):
    host = S3_AP_NORTHEAST2_HOST


class S3APNE2StorageDriver(S3StorageDriver):
    name = "Amazon S3 (ap-northeast-2)"
    connectionCls = S3APNE2Connection
    ex_location_name = "ap-northeast-2"
    region_name = "ap-northeast-2"


class S3APSouthConnection(S3SignatureV4Connection):
    host = S3_AP_SOUTH_HOST


class S3APSouthStorageDriver(S3StorageDriver):
    name = "Amazon S3 (ap-south-1)"
    connectionCls = S3APSouthConnection
    ex_location_name = "ap-south-1"
    region_name = "ap-south-1"


class S3SAEastConnection(S3SignatureV4Connection):
    host = S3_SA_EAST_HOST


class S3SAEastStorageDriver(S3StorageDriver):
    name = "Amazon S3 (sa-east-1)"
    connectionCls = S3SAEastConnection
    ex_location_name = "sa-east-1"
    region_name = "sa-east-1"


class S3CACentralConnection(S3SignatureV4Connection):
    host = S3_CA_CENTRAL_HOST


class S3CACentralStorageDriver(S3StorageDriver):
    name = "Amazon S3 (ca-central-1)"
    connectionCls = S3CACentralConnection
    ex_location_name = "ca-central-1"
    region_name = "ca-central-1"
